#! /usr/bin/env python
"""Validates commits in a CoLRev project."""
from __future__ import annotations

import datetime
import re
import typing
from pathlib import Path

from tqdm import tqdm

import colrev.exceptions as colrev_exceptions
import colrev.process.operation
import colrev.record.record
from colrev.constants import Fields
from colrev.constants import OperationsType
from colrev.constants import RecordState
from colrev.writer.write_utils import write_file


class Validate(colrev.process.operation.Operation):
    """Validate changes"""

    type = OperationsType.check

    def __init__(self, *, review_manager: colrev.review_manager.ReviewManager) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=self.type,
        )

        self.cpus = 4

    def _load_prior_records_dict(self, *, commit_sha: str) -> dict:
        """If commit is "": return the last commited version of records"""
        git_repo = self.review_manager.dataset.get_repo()
        # Ensure the path uses forward slashes, which is compatible with Git's path handling
        records_file_path = self.review_manager.paths.RECORDS_FILE_GIT
        revlist = (
            (
                commit_i.hexsha,
                (commit_i.tree / records_file_path).data_stream.read(),
            )
            for commit_i in git_repo.iter_commits(
                paths=str(self.review_manager.paths.RECORDS_FILE)
            )
        )

        found_target_commit = False
        for commit_id, filecontents in list(revlist):
            if commit_sha:
                if commit_id == commit_sha:
                    found_target_commit = True
                    continue
                if not found_target_commit:
                    continue
            elif not found_target_commit:
                # To skip the same commit
                found_target_commit = True
                continue
            prior_records_dict = colrev.loader.load_utils.loads(
                load_string=filecontents.decode("utf-8"),
                implementation="bib",
                logger=self.review_manager.logger,
            )

            return prior_records_dict
        return {}

    def _get_prep_prescreen_exclusions(self, records: dict) -> list:
        self.review_manager.logger.debug("Get prescreen exclusions...")

        target_commit = self._get_target_commit(scope="HEAD~1")
        prior_records_dict = self._load_prior_records_dict(commit_sha=target_commit)
        prep_prescreen_exclusions = []
        for record_dict in records.values():
            for prior_record_dict in prior_records_dict.values():
                if set(record_dict[Fields.ORIGIN]).intersection(
                    set(prior_record_dict[Fields.ORIGIN])
                ):
                    if (
                        prior_record_dict[Fields.STATUS]
                        != RecordState.rev_prescreen_excluded
                        and record_dict[Fields.STATUS]
                        == RecordState.rev_prescreen_excluded
                    ):
                        prep_prescreen_exclusions.append(record_dict)

        return prep_prescreen_exclusions

    def _get_change_diff(self, *, records: dict, origin_records: dict) -> list:
        self.review_manager.logger.debug("Calculating preparation differences...")

        change_diff = []
        for record_dict in records.values():

            origin_changes = []
            for origin in record_dict[Fields.ORIGIN]:
                if origin not in origin_records:
                    continue
                origin_record = origin_records[origin]

                change_score = colrev.record.record.Record.get_record_change_score(
                    colrev.record.record.Record(record_dict),
                    colrev.record.record.Record(origin_record),
                )

                origin_record["change_score"] = change_score
                origin_changes.append(origin_record)

            origin_changes.sort(key=lambda x: x["change_score"], reverse=False)

            change_diff.append(
                {
                    "record_dict": record_dict,
                    "change_score_max": (
                        max(
                            origin_change["change_score"]
                            for origin_change in origin_changes
                        )
                        if origin_changes
                        else 0
                    ),
                    "origins": origin_changes,
                }
            )

        # sort according to similarity
        change_diff.sort(key=lambda x: x["change_score_max"], reverse=True)
        return change_diff

    def _validate_prep_changes(self, *, report: dict) -> None:
        """Validate preparation changes"""

        self.review_manager.logger.info("Load records...")

        load_operation = self.review_manager.get_load_operation(
            hide_load_explanation=True
        )
        origin_records = {}
        for source in load_operation.load_active_sources(include_md=True):
            if not source.search_source.filename.is_file():
                continue
            load_operation.setup_source_for_load(source, select_new_records=False)
            for origin_record in source.search_source.source_records_list:
                origin_records[origin_record[Fields.ORIGIN][0]] = origin_record

        records = self.review_manager.dataset.load_records_dict()

        report["prep_prescreen_exclusions"] = self._get_prep_prescreen_exclusions(
            records
        )
        report["prep"] = self._get_change_diff(
            records=records, origin_records=origin_records
        )

    def _export_merge_candidates_file(self, records: list[dict]) -> None:
        merge_candidates_file = Path("data/dedupe/merge_candidates_file.txt")
        merge_candidates_file.parent.mkdir(exist_ok=True, parents=True)

        with open(merge_candidates_file, "w", encoding="utf-8") as file:
            for ref_rec_dict in tqdm(records):
                ref_rec = colrev.record.record.Record(ref_rec_dict)
                for comp_rec_dict in reversed(records):
                    # Note : due to symmetry, we only need one part of the matrix
                    if ref_rec_dict[Fields.ID] == comp_rec_dict[Fields.ID]:
                        break
                    comp_rec = colrev.record.record.Record(comp_rec_dict)
                    similarity = colrev.record.record.Record.get_record_similarity(
                        ref_rec, comp_rec
                    )

                    if similarity > 0.95:
                        print(f"{ref_rec_dict['ID']}-{comp_rec_dict['ID']}")

                        file.write(ref_rec.format_bib_style())
                        file.write("\n")
                        file.write(comp_rec.format_bib_style())
                        file.write("\n")
                        file.write(
                            f"colrev dedupe -m {ref_rec_dict['ID']},{comp_rec_dict['ID']}\n\n"
                        )

        if merge_candidates_file.read_text(encoding="utf-8") == "":
            merge_candidates_file.unlink()

    def _validate_dedupe_changes(self, *, report: dict, commit_sha: str) -> None:
        """Validate dedupe changes"""

        # at some point, we may allow users to validate
        # all duplicates/non-duplicates (across commits)

        # if self._gids_conflict(main_record=main_record, dupe_record=dupe_record):
        #     self.review_manager.logger.info(
        #         "Prevented merge with conflicting global IDs: "
        #         f"{main_record.data[Fields.ID]} - {dupe_record.data[Fields.ID]}"
        #     )
        #     return True

        # def _gids_conflict(
        #     self, *, main_record: colrev.record.record.Record,
        #               sdupe_record: colrev.record.record.Record
        # ) -> bool:
        #     gid_conflict = False
        #     if Fields.DOI in main_record.data and Fields.DOI in dupe_record.data:
        #         doi_main = main_record.data.get(Fields.DOI, "a").replace("\\", "")
        #         doi_dupe = dupe_record.data.get(Fields.DOI, "b").replace("\\", "")
        #         if doi_main != doi_dupe:
        #             gid_conflict = True

        #     return gid_conflict

        records = self._load_changed_records(commit_sha=commit_sha)

        prior_records_dict = self._load_prior_records_dict(commit_sha=commit_sha)
        # Note : the if-statement avoids time-consuming procedures when the
        # origin-sets have not changed (no duplicates merged)
        if not self._deduplicated_records(
            records=records, prior_records_dict=prior_records_dict
        ):
            report["dedupe"] = []
            return

        change_diff = []
        merged_records = False
        for record in records:
            if "changed_in_target_commit" not in record:
                continue
            del record["changed_in_target_commit"]

            if len(record[Fields.ORIGIN]) == 1:
                continue
            merged_records = True

            merged_records_list = []

            for prior_record in prior_records_dict.values():
                if len(prior_record[Fields.ORIGIN]) == 1:
                    continue
                if any(o in record[Fields.ORIGIN] for o in prior_record[Fields.ORIGIN]):
                    merged_records_list.append(prior_record)

            if len(merged_records_list) < 2:
                # merged records not found
                continue

            reference_record = merged_records_list.pop(0)
            # Note : should usually be only one merged_rec (but multiple-merges are possible)
            for merged_rec in merged_records_list:
                change_score = colrev.record.record.Record.get_record_change_score(
                    colrev.record.record.Record(reference_record),
                    colrev.record.record.Record(merged_rec),
                )
                change_diff.append(
                    {
                        "record": record,
                        "prior_record_a": reference_record,
                        "prior_record_b": merged_rec,
                        "change_score": change_score,
                    }
                )

        change_diff = [
            element for element in change_diff if element["change_score"] < 1
        ]
        if 0 == len(change_diff):
            if merged_records:
                self.review_manager.logger.info("No substantial differences found.")
            else:
                self.review_manager.logger.info("No merged records")

        self._export_merge_candidates_file(records)

        # sort according to similarity
        change_diff.sort(key=lambda x: x["change_score"], reverse=True)

        report["dedupe"] = change_diff

    def _get_changed_records(self, *, target_commit: str) -> typing.List[dict]:
        """Get the records that changed in a selected commit"""

        dataset = self.review_manager.dataset
        git_repo = dataset.get_repo()
        revlist = (
            (
                commit.hexsha,
                (
                    commit.tree / self.review_manager.paths.RECORDS_FILE_GIT
                ).data_stream.read(),
            )
            for commit in git_repo.iter_commits(
                paths=str(self.review_manager.paths.RECORDS_FILE)
            )
        )
        found = False
        records: typing.Dict[str, typing.Any] = {}
        prior_records = {}
        for commit, filecontents in list(revlist):
            if found:  # load the records_file_relative in the following commit
                prior_records = colrev.loader.load_utils.loads(
                    load_string=filecontents.decode("utf-8"),
                    implementation="bib",
                    logger=self.review_manager.logger,
                )
                break
            if commit == target_commit:
                records = colrev.loader.load_utils.loads(
                    load_string=filecontents.decode("utf-8"),
                    implementation="bib",
                    logger=self.review_manager.logger,
                )
                found = True

        # determine which records have been changed (prepared or merged)
        # in the target_commit
        for record in records.values():
            prior_record_l = [
                rec
                for rec in prior_records.values()
                if any(x in record[Fields.ORIGIN] for x in rec[Fields.ORIGIN])
            ]
            if prior_record_l:
                prior_record = prior_record_l[0]
                # Note: the following is an exact comparison of all fields
                if record != prior_record:
                    record.update(changed_in_target_commit="True")

        return list(records.values())

    def _load_changed_records(
        self, *, commit_sha: typing.Optional[str] = None
    ) -> list[dict]:
        """Load the records that were changed in the target commit"""
        if commit_sha is None:
            self.review_manager.logger.info("Loading data...")
            records = self.review_manager.dataset.load_records_dict()
            for record_dict in records.values():
                record_dict.update(changed_in_target_commit="True")
            return list(records.values())

        self.review_manager.logger.info("Loading data from history...")
        changed_records = self._get_changed_records(target_commit=commit_sha)

        return changed_records

    def _validate_properties(self, *, commit_sha: str) -> dict:
        """Validate properties"""

        # option: --history: check all preceding commits (create a list...)

        git_repo = self.review_manager.dataset.get_repo()

        cur_sha = git_repo.head.commit.hexsha
        cur_branch = git_repo.active_branch.name
        report: typing.Dict[str, typing.Any] = {}

        if git_repo.is_dirty() and not commit_sha == cur_sha:
            self.review_manager.logger.error(
                "Error: Need a clean repository to validate properties "
                "of prior commit"
            )
            return {}

        if not commit_sha == cur_sha:
            self.review_manager.logger.info(f"Check out target_commit = {commit_sha}")
            git_repo.git.checkout(commit_sha)

        ret = self.review_manager.check_repo()
        if 0 == ret["status"]:
            report["record_traceability"] = True
            report["consistency"] = True

        else:
            report["record_traceability"] = False
            report["consistency"] = False

        completeness_condition = self.review_manager.get_completeness_condition()
        if completeness_condition:
            report["completeness"] = True

        else:
            report["completeness"] = False

        git_repo.git.checkout(cur_branch, force=True)

        return {"properties": report}

    def _validate_general(self, *, commit_sha: str) -> dict:
        return {
            "general": {
                "commit": commit_sha,
                "commit_relative": self._get_relative_commit(commit_sha=commit_sha),
            }
        }

    def _set_scope_based_on_target_commit(self, *, commit_sha: str) -> str:
        # pylint: disable=too-many-branches

        if not commit_sha:
            commit_sha = self.review_manager.dataset.get_last_commit_sha()

        git_repo = self.review_manager.dataset.get_repo()

        revlist = list(
            (
                commit.hexsha,
                commit.message,
            )
            for commit in git_repo.iter_commits()
        )

        scope = ""
        # Note : simple heuristic: commit messages
        for commit_id, msg in revlist:
            if commit_id == commit_sha:
                if "colrev prep" in msg:
                    scope = "prepare"
                elif "colrev dedupe" in msg:
                    # pylint: disable=colrev-missed-constant-usage
                    scope = "dedupe"
                elif any(
                    x in msg
                    for x in [
                        "colrev init",
                        "colrev load",
                        "colrev pdf-get",
                        "colrev pdf-prep",
                        "colrev screen",
                        "colrev prescreen",
                    ]
                ):
                    scope = "general"
                else:
                    scope = "general"
            if scope != "":
                break

        # Otherwise: compare records
        if scope in ["general"]:
            # detect transition types in the respective commit and
            # use them to calculate the report
            records: typing.Dict[str, typing.Dict] = {}
            hist_records: typing.Dict[str, typing.Dict] = {}
            for recs in self.review_manager.dataset.load_records_from_history(
                commit_sha
            ):
                # continue
                if not records:
                    records = recs
                    continue
                if not hist_records:
                    hist_records = recs
                    break

            # Note : still very simple heuristics...
            if len(records) != len(hist_records):
                # pylint: disable=colrev-missed-constant-usage
                scope = "dedupe"

        return scope

    def _get_filter_setting(
        self, *, filter_setting: str, properties: bool, commit_sha: str, scope: str
    ) -> str:
        if self._is_contributor_validation_condition(scope):
            filter_setting = "contributor"

        elif properties:
            filter_setting = "properties"
        elif filter_setting == "all":
            filter_setting = self._set_scope_based_on_target_commit(
                commit_sha=commit_sha
            )

        self.review_manager.logger.info(f"Filter: {filter_setting} changes")
        return filter_setting

    def validate_merge_prescreen_screen(
        self,
        *,
        commit_sha: str = "CURRENT",
        current_branch_records: dict,
        other_branch_records: dict,
        records_reconciled: dict,
        # current_branch_name: str = "",
        # other_branch_name: str = "",
    ) -> dict:
        """Validate presceen/screen between merged branches"""

        prescreen_validation = {"commit_sha": commit_sha, "prescreen": []}
        for rid, record_dict in current_branch_records.items():
            prescreen_validation["prescreen"].append(  # type: ignore
                {
                    Fields.ID: rid,
                    "coder1": str(record_dict[Fields.STATUS]),
                    "coder2": str(other_branch_records[rid][Fields.STATUS]),
                    "reconciled": str(records_reconciled[rid][Fields.STATUS]),
                }
            )
        prescreen_validation["statistics"] = {  # type: ignore
            "percentage_agreement": len(
                [
                    x
                    for x in prescreen_validation["prescreen"]
                    if x["coder1"] == x["coder2"]  # type: ignore
                ]
            )
            / len(prescreen_validation["prescreen"])
        }
        return prescreen_validation

    def _validate_merge_changes(self) -> dict:
        """Validate merge changes (reconciliation between branches)"""

        merge_validation = []

        git_repo = self.review_manager.dataset.get_repo()

        revlist = git_repo.iter_commits(
            paths=str(self.review_manager.paths.RECORDS_FILE)
        )
        # Ensure the path uses forward slashes, which is compatible with Git's path handling

        for commit in list(revlist):
            if len(commit.parents) <= 1:
                continue

            if not any(x in commit.message for x in ["prescreen", "screen"]):
                continue

            load_str = (
                (commit.parents[0].tree / self.review_manager.paths.RECORDS_FILE_GIT)
                .data_stream.read()
                .decode("utf-8")
            )
            records_branch_1 = colrev.loader.load_utils.loads(
                load_string=load_str,
                implementation="bib",
                logger=self.review_manager.logger,
            )

            load_str = (
                (commit.parents[1].tree / self.review_manager.paths.RECORDS_FILE_GIT)
                .data_stream.read()
                .decode("utf-8")
            )
            records_branch_2 = colrev.loader.load_utils.loads(
                load_string=load_str,
                implementation="bib",
                logger=self.review_manager.logger,
            )

            load_str = (
                (commit.tree / self.review_manager.paths.RECORDS_FILE_GIT)
                .data_stream.read()
                .decode("utf-8")
            )
            records_reconciled = colrev.loader.load_utils.loads(
                load_string=load_str,
                implementation="bib",
                logger=self.review_manager.logger,
            )

            if "screen" in commit.message or "prescreen" in commit.message:
                prescreen_validation = self.validate_merge_prescreen_screen(
                    commit_sha=commit.hexsha,
                    current_branch_records=records_branch_1,
                    other_branch_records=records_branch_2,
                    records_reconciled=records_reconciled,
                )
                merge_validation.append(prescreen_validation)

        return {"merge": merge_validation}

    def _get_target_commit(self, *, scope: str, filter_setting: str = "") -> str:
        """Get the commit from commit sha or tree hash"""

        commit = ""
        if filter_setting == "contributor":
            return commit

        git_repo = self.review_manager.dataset.get_repo()
        if scope in ["HEAD", "."]:
            scope = "HEAD~0"
        if scope.startswith("HEAD~"):
            assert scope.replace("HEAD~", "").isdigit()
            back_count = int(scope.replace("HEAD~", ""))
            for commit_item in git_repo.iter_commits():
                if back_count == 0:
                    commit = commit_item.hexsha
                    break
                back_count -= 1
        else:
            valid_options = []

            try:
                commit_object = git_repo.commit(scope)
                commit = commit_object.hexsha
            except ValueError as exc:
                for commit_candidate in git_repo.iter_commits():
                    if str(commit_candidate.tree) == scope:
                        commit = commit_candidate.hexsha
                        break
                    valid_options.append(str(commit_candidate.tree))

                if commit == "":
                    # pylint: disable=raise-missing-from
                    raise colrev_exceptions.ParameterError(
                        parameter="validate.scope", value=scope, options=valid_options
                    ) from exc

        if not re.match(r"[0-9a-f]{5,40}", commit):
            raise colrev_exceptions.ParameterError(
                parameter="commit",
                value=scope,
                options=[x.hexsha for x in git_repo.iter_commits()],
            )
        return commit

    def _deduplicated_records(
        self, *, records: list[dict], prior_records_dict: dict
    ) -> bool:
        return {",".join(sorted(x)) for x in [r[Fields.ORIGIN] for r in records]} != {
            ",".join(sorted(x))
            for x in [r[Fields.ORIGIN] for r in prior_records_dict.values()]
        }

    def _get_contributor_validation(self, *, scope: str) -> dict:
        report: typing.Dict[str, typing.Any] = {"contributor_commits": []}
        valid_options = []
        git_repo = self.review_manager.dataset.get_repo()
        for commit in git_repo.iter_commits():
            if any(
                x == scope
                for x in [
                    commit.author.email,
                    commit.author.name,
                    commit.committer.email,
                    commit.committer.name,
                ]
            ):
                commit_date = datetime.datetime.fromtimestamp(commit.committed_date)
                report["contributor_commits"].append(
                    {
                        "msg": commit.message.split("\n", maxsplit=1)[0],
                        "date": commit_date,
                        "author": commit.author.name,
                        "author_email": commit.author.email,
                        "commit_sha": commit.hexsha,
                        "committer": commit.committer.name,
                        "committer_email": commit.committer.email,
                        "validate": f"colrev validate {commit.hexsha}",
                    }
                )

            if not self.review_manager.verbose_mode:
                if "script" in commit.author.name:
                    continue
            if commit.author.name not in valid_options:
                valid_options.append(commit.author.name)
            if commit.author.email not in valid_options:
                valid_options.append(commit.author.email)

        if not report["contributor_commits"]:
            raise colrev_exceptions.ParameterError(
                parameter="validate.contributor",
                value=scope,
                options=valid_options,
            )
        return report

    def _get_relative_commit(self, commit_sha: str) -> str:
        git_repo = self.review_manager.dataset.get_repo()

        relative_to_head = 0
        for commit_i in git_repo.iter_commits():
            if commit_sha == commit_i.hexsha:
                return f"HEAD~{relative_to_head}"
            relative_to_head += 1

        return commit_sha

    def _is_contributor_validation_condition(self, scope: str) -> bool:
        return (
            "HEAD" not in scope
            and not re.match(r"[0-9a-f]{5,40}", scope)
            and "." != scope
        )

    def remove_md_origins(self, origins_to_remove: dict) -> None:
        """Remove origin records for md_ files"""

        records = self.review_manager.dataset.load_records_dict()
        origins_in_use = [
            o for record in records.values() for o in record[Fields.ORIGIN]
        ]

        for file, identifiers in origins_to_remove.items():
            if file.startswith("md_"):
                filename = self.review_manager.paths.SEARCH_DIR / Path(file)
                origin_records = colrev.loader.load_utils.load(
                    filename=filename,
                    logger=self.review_manager.logger,
                )
                for identifier in identifiers:
                    if f"{file}/{identifier}" in origins_in_use:
                        continue
                    if identifier in origin_records:
                        origin_records.pop(identifier)
                write_file(records_dict=origin_records, filename=filename)

            self.review_manager.dataset.add_changes(filename)

    @colrev.process.operation.Operation.decorate()
    def main(
        self,
        *,
        scope: str,
        filter_setting: str,
        properties: bool = False,
    ) -> dict:
        """Validate a commit (main entrypoint)"""

        target_commit = self._get_target_commit(
            scope=scope, filter_setting=filter_setting
        )

        filter_setting = self._get_filter_setting(
            filter_setting=filter_setting,
            properties=properties,
            commit_sha=target_commit,
            scope=scope,
        )

        if filter_setting == "general":
            return self._validate_general(commit_sha=target_commit)

        if filter_setting == "contributor":
            return self._get_contributor_validation(scope=scope)

        if filter_setting == "properties":
            return self._validate_properties(commit_sha=target_commit)

        if filter_setting == "merge":  # for git branches (not colrev dedupe)
            return self._validate_merge_changes()

        report: typing.Dict[str, typing.Any] = {}
        if filter_setting in ["prepare", "all"]:
            self._validate_prep_changes(report=report)
        if filter_setting in ["dedupe", "all"]:
            self._validate_dedupe_changes(report=report, commit_sha=target_commit)
        return report



#! /usr/bin/env python
"""Quality model for records."""
from __future__ import annotations

import importlib
import typing
from multiprocessing import Lock
from pathlib import Path

import colrev.record.qm.checkers
import colrev.record.record
from colrev.constants import Fields


class QualityModel:
    """The quality model for records"""

    checkers = []  # type: ignore

    def __init__(
        self,
        *,
        defects_to_ignore: list[str],
        pdf_mode: bool = False,
        path: typing.Optional[Path] = None,
    ) -> None:
        self.pdf_mode = pdf_mode
        self.defects_to_ignore = defects_to_ignore
        self._register_checkers()
        self.local_index_lock = Lock()
        self.path = path

    def _register_checkers(self) -> None:
        """Register checkers from the checker directory, looking for a
        'register' function in each one.
        """

        self.checkers = []
        if self.pdf_mode:
            module_path = "colrev.record.qm.pdf_checkers."
        else:
            module_path = "colrev.record.qm.checkers."

        if self.pdf_mode:
            checker_path = Path(__file__).parent / Path("pdf_checkers/")
        else:
            checker_path = Path(__file__).parent / Path("checkers")

        for filename in checker_path.glob("*.py"):
            if "__init__" in str(filename):
                continue

            try:
                module = importlib.import_module(module_path + filename.stem)
            except ValueError as exc:  # pragma: no cover
                print(f"Problem with filepath for module import {filename}: {exc}")
            except ImportError as exc:  # pragma: no cover
                print(f"Problem importing module {filename}: {exc}")
            else:
                if hasattr(module, "register"):
                    module.register(self)
                else:  # pragma: no cover
                    print(f"Module {filename} does not have a register function")

    def register_checker(self, checker) -> None:  # type: ignore
        """Register a checker"""
        self.checkers.append(checker)

    def run(self, *, record: colrev.record.record.Record) -> None:
        """Run the checkers"""

        if self.pdf_mode:
            if not self.path:
                raise ValueError("Path is required for PDF mode")
            if (
                Fields.FILE not in record.data
                or not Path(record.data[Fields.FILE]).is_file()
            ):
                return
            # text_from_pdf is already set in tests
            if (
                Fields.TEXT_FROM_PDF not in record.data
                or Fields.NR_PAGES_IN_FILE not in record.data
            ):
                # The following should be improved.
                record = colrev.record.record_pdf.PDFRecord(record.data, path=self.path)
                record.set_text_from_pdf()

        for checker in self.checkers:
            if checker.msg in self.defects_to_ignore:
                continue
            checker.run(record=record)

        if self.pdf_mode:
            record.data.pop(Fields.TEXT_FROM_PDF, None)
            record.data.pop(Fields.NR_PAGES_IN_FILE, None)



#! /usr/bin/env python
"""Load utils for AIS"""
from __future__ import annotations

from colrev.constants import ENTRYTYPES
from colrev.constants import Fields


def enl_id_labeler(records: list) -> None:
    """Labeler for IDs in ENL files."""
    for record_dict in records:
        record_dict[Fields.ID] = record_dict["U"].replace(
            "https://aisel.aisnet.org/", ""
        )


def enl_entrytype_setter(record_dict: dict) -> None:
    """Set the entrytype for ENL files."""
    if "0" not in record_dict:
        keys_to_check = ["V", "N"]
        if any(k in record_dict for k in keys_to_check):
            record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.ARTICLE
        else:
            record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.INPROCEEDINGS
    else:
        if record_dict["0"] == "Journal Article":
            record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.ARTICLE
        elif record_dict["0"] == "Inproceedings":
            record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.INPROCEEDINGS
        else:
            record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.MISC


def bib_field_mapper(record_dict: dict) -> None:
    """Map the fields for BibTeX files."""

    record_dict.pop("type", None)
    record_dict.pop("IG", None)

    record_dict.pop("H", None)
    record_dict.pop("NT", None)
    record_dict.pop("OO", None)
    record_dict.pop("OU", None)
    record_dict.pop("ND", None)
    record_dict.pop("pe", None)
    record_dict.pop("ns", None)


def enl_field_mapper(record_dict: dict) -> None:
    """Map the fields for ENL files."""
    key_maps = {
        ENTRYTYPES.ARTICLE: {
            "T": Fields.TITLE,
            "A": Fields.AUTHOR,
            "D": Fields.YEAR,
            "B": Fields.JOURNAL,
            "V": Fields.VOLUME,
            "N": Fields.NUMBER,
            "P": Fields.PAGES,
            "X": Fields.ABSTRACT,
            "U": Fields.URL,
            "8": "date",
            "0": "type",
        },
        ENTRYTYPES.INPROCEEDINGS: {
            "T": Fields.TITLE,
            "A": Fields.AUTHOR,
            "D": Fields.YEAR,
            "B": Fields.BOOKTITLE,
            "V": Fields.VOLUME,
            "N": Fields.NUMBER,
            "P": Fields.PAGES,
            "X": Fields.ABSTRACT,
            "U": Fields.URL,
            "8": "date",
            "0": "type",
        },
    }

    key_map = key_maps[record_dict[Fields.ENTRYTYPE]]
    for ris_key in list(record_dict.keys()):
        if ris_key in key_map:
            standard_key = key_map[ris_key]
            record_dict[standard_key] = record_dict.pop(ris_key)

    # Add secondary titles / fix cases where F and P fields start with a space:
    # these may be special cases:
    # https://aisel.aisnet.org/amcis2002/301/
    # https://aisel.aisnet.org/amcis2002/145/
    if " F" in record_dict:
        record_dict[Fields.TITLE] = (
            record_dict[Fields.TITLE] + " " + record_dict.pop(" F")
        )
    if " P" in record_dict:
        record_dict[Fields.TITLE] = (
            record_dict[Fields.TITLE] + " " + record_dict.pop(" P")
        )

    if Fields.AUTHOR in record_dict and isinstance(record_dict[Fields.AUTHOR], list):
        record_dict[Fields.AUTHOR] = " and ".join(record_dict[Fields.AUTHOR])
    if Fields.EDITOR in record_dict and isinstance(record_dict[Fields.EDITOR], list):
        record_dict[Fields.EDITOR] = " and ".join(record_dict[Fields.EDITOR])
    if Fields.KEYWORDS in record_dict and isinstance(
        record_dict[Fields.KEYWORDS], list
    ):
        record_dict[Fields.KEYWORDS] = ", ".join(record_dict[Fields.KEYWORDS])

    record_dict.pop("type", None)

    for key, value in record_dict.items():
        record_dict[key] = str(value)



#! /usr/bin/env python
"""Discovering and using packages."""
from __future__ import annotations

import typing
from pathlib import Path

from pydantic import BaseModel
from pydantic import Field

import colrev.process.operation
import colrev.record.record
import colrev.settings
from colrev.constants import SearchType


class DefaultSettings(BaseModel):
    """Endpoint settings"""

    endpoint: str = Field()


class DefaultSourceSettings(colrev.settings.SearchSource, BaseModel):
    """Search source settings"""

    # pylint: disable=duplicate-code
    # pylint: disable=too-many-instance-attributes
    endpoint: str
    filename: Path
    search_type: SearchType
    search_parameters: dict
    comment: typing.Optional[str]



#! /usr/bin/env python
"""Convenience functions for load formatting"""
from __future__ import annotations

import html
import re

import colrev.env.language_service
import colrev.exceptions as colrev_exceptions
from colrev.constants import Fields
from colrev.constants import FieldValues
from colrev.constants import RecordState

# pylint: disable=too-few-public-methods


class LoadFormatter:
    """Load formatter class"""

    # Based on
    # https://en.wikibooks.org/wiki/LaTeX/Special_Characters
    _LATEX_SPECIAL_CHAR_MAPPING = {
        '\\"a': "ä",
        '\\"o': "ö",
        '\\"u': "ü",
        '\\"A': "Ä",
        '\\"O': "Ö",
        '\\"U': "Ü",
        "\\&": "&",
        "\\textendash": "–",
        "\\textemdash": "—",
        "\\~a": "ã",
        "\\'o": "ó",
        "\\emph": "",
        "\\textit": "",
        "\\'e": "é",
        "\\`e": "è",
        '"a': "ä",
        '"o': "ö",
        '"u': "ü",
    }

    _FIELDS_TO_PROCESS = [
        Fields.AUTHOR,
        Fields.YEAR,
        Fields.TITLE,
        Fields.JOURNAL,
        Fields.BOOKTITLE,
        Fields.SERIES,
        Fields.VOLUME,
        Fields.NUMBER,
        Fields.PAGES,
        Fields.DOI,
        Fields.ABSTRACT,
    ]

    def __init__(self) -> None:
        self.language_service = colrev.env.language_service.LanguageService()

    def _fix_author_particles(self, record: colrev.record.record.Record) -> None:
        # Fix the name particles in the author field
        if Fields.AUTHOR in record.data:
            names = record.data[Fields.AUTHOR].split(" and ")
            for ind, name in enumerate(names):
                for prefix in [
                    "van den",
                    "von den",
                    "van der",
                    "von der",
                    "vom",
                    "van",
                    "von",
                ]:
                    if name.startswith(f"{prefix} "):
                        if "," in name:
                            name = "{" + name.replace(", ", "}, ")
                        else:
                            name = "{" + name + "}"
                    if name.endswith(f" {prefix}"):
                        if "," in name:
                            name = (
                                "{"
                                + prefix
                                + " "
                                + name[: -len(prefix)].rstrip().replace(", ", "}, ")
                            )
                        else:
                            name = (
                                "{" + prefix + " " + name[: -len(prefix)].rstrip() + "}"
                            )

                    names[ind] = name
            record.data[Fields.AUTHOR] = " and ".join(names)

    def _format_doi(self, record: colrev.record.record.Record) -> None:
        if Fields.DOI in record.data:
            record.data[Fields.DOI] = (
                record.data[Fields.DOI]
                .lower()
                .replace("https://", "http://")
                .replace("dx.doi.org", "doi.org")
                .replace("http://doi.org/", "")
                .upper()
            )

    def _unify_language(self, record: colrev.record.record.Record) -> None:
        if Fields.LANGUAGE in record.data and len(record.data[Fields.LANGUAGE]) != 3:
            try:
                self.language_service.unify_to_iso_639_3_language_codes(record=record)
            except colrev_exceptions.InvalidLanguageCodeException:
                del record.data[Fields.LANGUAGE]

    def _rename_issue_to_number(self, record: colrev.record.record.Record) -> None:
        if Fields.NUMBER not in record.data and "issue" in record.data:
            record.data[Fields.NUMBER] = record.data.pop("issue")

    def _apply_strict_requirements(
        self, *, record: colrev.record.record.Record
    ) -> None:

        self._fix_author_particles(record)
        self._format_doi(record)
        self._unify_language(record)
        self._rename_issue_to_number(record)

    def _unescape_latex(self, *, input_str: str) -> str:
        for latex_char, repl_char in self._LATEX_SPECIAL_CHAR_MAPPING.items():
            input_str = input_str.replace(f"{{{latex_char}}}", repl_char)
            input_str = input_str.replace(latex_char, repl_char)
        return input_str

    def _unescape_html(self, *, input_str: str) -> str:
        input_str = html.unescape(input_str)
        if "<" in input_str:
            input_str = re.sub(r"<.*?>", "", input_str)
        return input_str

    def _unescape_field_values(self, *, record: colrev.record.record.Record) -> None:
        for field in record.data:
            if field not in self._FIELDS_TO_PROCESS:
                continue
            record.data[field] = str(record.data[field])
            if "\\" in record.data[field]:
                record.data[field] = self._unescape_latex(input_str=record.data[field])
            record.data[field] = self._unescape_html(input_str=record.data[field])

            record.data[field] = record.data[field].replace("\n", " ").rstrip().lstrip()

    def _standardize_field_values(self, *, record: colrev.record.record.Record) -> None:
        if record.data.get(Fields.TITLE, FieldValues.UNKNOWN) != FieldValues.UNKNOWN:
            record.data[Fields.TITLE] = re.sub(
                r"\s+", " ", record.data[Fields.TITLE]
            ).rstrip(".")

        # Fix floating point years
        if Fields.YEAR in record.data and str(record.data[Fields.YEAR]).endswith(".0"):
            record.data[Fields.YEAR] = str(record.data[Fields.YEAR])[:-2]

        if Fields.PAGES in record.data:
            record.data[Fields.PAGES] = record.data[Fields.PAGES].replace("–", "--")
            if record.data[Fields.PAGES].count("-") == 1:
                record.data[Fields.PAGES] = record.data[Fields.PAGES].replace("-", "--")
            if record.data[Fields.PAGES].lower() == "n.pag":
                del record.data[Fields.PAGES]

        if record.data.get(Fields.VOLUME, "") == "ahead-of-print":
            del record.data[Fields.VOLUME]
        if record.data.get(Fields.NUMBER, "") == "ahead-of-print":
            del record.data[Fields.NUMBER]

        if Fields.URL in record.data and "login?url=https" in record.data[Fields.URL]:
            record.data[Fields.URL] = record.data[Fields.URL][
                record.data[Fields.URL].find("login?url=https") + 10 :
            ]

    def run(self, record: colrev.record.record.Record) -> None:
        """Run the load formatter"""

        self._apply_strict_requirements(record=record)

        if (
            Fields.STATUS in record.data
            and record.data[Fields.STATUS] != RecordState.md_retrieved
        ):
            return

        self._unescape_field_values(record=record)
        self._standardize_field_values(record=record)



#! /usr/bin/env python
"""SearchSource: Taylor and Francis"""
from __future__ import annotations

import re
from pathlib import Path

import zope.interface
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class TaylorAndFrancisSearchSource:
    """Taylor and Francis"""

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings
    endpoint = "colrev.taylor_and_francis"
    source_identifier = "{{doi}}"
    search_types = [SearchType.DB]
    ci_supported: bool = Field(default=False)
    heuristic_status = SearchSourceHeuristicStatus.supported

    def __init__(
        self, *, source_operation: colrev.process.operation.Operation, settings: dict
    ) -> None:
        self.search_source = self.settings_class(**settings)
        self.review_manager = source_operation.review_manager
        self.source_operation = source_operation

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for Taylor and Francis"""

        result = {"confidence": 0.0}

        if data.count("\n@") > 1:
            if data.count("eprint = { \n    \n    ") >= data.count("\n@"):
                result["confidence"] = 1.0

        return result

    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint (based on query provided to colrev search --add )"""

        search_source = operation.create_db_source(
            search_source_cls=cls,
            params={},
        )
        operation.add_source_and_search(search_source)
        return search_source

    def search(self, rerun: bool) -> None:
        """Run a search of TaylorAndFrancis"""

        if self.search_source.search_type == SearchType.DB:
            self.source_operation.run_db_search(  # type: ignore
                search_source_cls=self.__class__,
                source=self.search_source,
            )
            return

        raise NotImplementedError

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.record.Record:
        """Not implemented"""
        return record

    def _load_bib(self) -> dict:
        def field_mapper(record_dict: dict) -> None:
            if "note" in record_dict:
                record_dict[f"{self.endpoint}.note"] = record_dict.pop("note")
            if "eprint" in record_dict:
                record_dict[f"{self.endpoint}.eprint"] = record_dict.pop("eprint")

            for key in list(record_dict.keys()):
                if key not in ["ID", "ENTRYTYPE"]:
                    record_dict[key.lower()] = record_dict.pop(key)

        records = colrev.loader.load_utils.load(
            filename=self.search_source.filename,
            logger=self.review_manager.logger,
            unique_id_field="ID",
            field_mapper=field_mapper,
        )
        return records

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            return self._load_bib()

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.record.Record:
        """Source-specific preparation for Taylor and Francis"""

        # remove eprint and URL fields (they only have dois...)
        record.remove_field(key="colrev.taylor_and_francis.eprint")
        if "colrev.taylor_and_francis.note" in record.data and re.match(
            r"PMID: \d*", record.data["colrev.taylor_and_francis.note"]
        ):
            record.rename_field(
                key="colrev.taylor_and_francis.note", new_key=Fields.PUBMED_ID
            )
            record.data[Fields.PUBMED_ID] = record.data[Fields.PUBMED_ID][6:]

        return record



#! /usr/bin/env python
"""Manages environment registry, services, and stauts"""
from __future__ import annotations

import json
import logging
import typing
from pathlib import Path

import git
import yaml

import colrev.exceptions as colrev_exceptions
import colrev.ops.check
import colrev.process.operation
import colrev.record.record
from colrev.constants import Fields
from colrev.constants import FieldValues
from colrev.constants import Filepaths
from colrev.env.utils import dict_set_nested
from colrev.env.utils import get_by_path


class EnvironmentManager:
    """The EnvironmentManager manages environment resources and services"""

    load_yaml = False

    def __init__(self) -> None:
        self.environment_registry = self.load_environment_registry()
        self._registered_ports: typing.List[str] = []

    def register_ports(self, ports: typing.List[str]) -> None:
        """Register a localhost port to avoid conflicts"""
        for port_to_register in ports:
            if port_to_register in self._registered_ports:
                raise colrev_exceptions.PortAlreadyRegisteredException(
                    f"Port {port_to_register} already registered"
                )
            self._registered_ports.append(port_to_register)

    def load_environment_registry(self) -> dict:
        """Load the local registry"""
        environment_registry = {}
        if Filepaths.REGISTRY_FILE.is_file():
            self.load_yaml = False
            with open(Filepaths.REGISTRY_FILE, encoding="utf8") as file:
                environment_registry = json.load(fp=file)
            # assert "local_index" in environment_registry
            # assert "packages" in environment_registry

        return environment_registry

    def local_repos(self) -> list:
        """gets local repos from local index"""
        self.environment_registry = self.load_environment_registry()
        if (
            "local_index" not in self.environment_registry
            or "repos" not in self.environment_registry["local_index"]
        ):
            return []
        return self.environment_registry["local_index"]["repos"]

    def _cast_values_to_str(self, data) -> dict:  # type: ignore
        result = {}
        for key, value in data.items():
            if isinstance(value, dict):
                result[key] = self._cast_values_to_str(value)
            elif isinstance(value, list):
                result[key] = [self._cast_values_to_str(v) for v in value]  # type: ignore
            else:
                result[key] = str(value)  # type: ignore
        return result

    def save_environment_registry(self, updated_registry: dict) -> None:
        """Save the local registry"""
        Filepaths.REGISTRY_FILE.parents[0].mkdir(parents=True, exist_ok=True)
        with open(Filepaths.REGISTRY_FILE, "w", encoding="utf8") as file:
            json.dump(
                dict(self._cast_values_to_str(updated_registry)), indent=4, fp=file
            )

    def register_repo(
        self,
        path_to_register: Path,
        logger: logging.Logger = logging.getLogger(__name__),
    ) -> None:
        """Register a repository"""
        path_to_register = path_to_register.resolve().absolute()
        self.environment_registry = self.load_environment_registry()

        if "local_index" not in self.environment_registry:
            self.environment_registry["local_index"] = {"repos": []}
        registered_paths = [
            x["repo_source_path"]
            for x in self.environment_registry["local_index"]["repos"]
        ]

        if registered_paths:
            if str(path_to_register) in registered_paths:
                logger.warning(f"Path already registered: {path_to_register}")
                return
        else:
            logger.info("Register %s in %s", path_to_register, Filepaths.REGISTRY_FILE)

        new_record = {
            "repo_name": path_to_register.stem,
            "repo_source_path": path_to_register,
        }
        git_repo = git.Repo(path_to_register)
        try:
            remote_urls = list(git_repo.remote("origin").urls)
            new_record["repo_source_url"] = remote_urls[0]
        except (ValueError, IndexError):  # pragma: no cover
            for remote in git_repo.remotes:
                if remote.url:
                    new_record["repo_source_url"] = remote.url
                    break
        self.environment_registry["local_index"]["repos"].append(new_record)
        self.save_environment_registry(self.environment_registry)
        logger.info("Registered path (%s)", path_to_register)

    @classmethod
    def get_name_mail_from_git(cls) -> typing.Tuple[str, str]:  # pragma: no cover
        """Get the committer name and email from git (globals)"""
        global_conf_details = ("NA", "NA")
        try:
            username = git.config.GitConfigParser().get_value("user", "name")
            email = git.config.GitConfigParser().get_value("user", "email")
            global_conf_details = (username, email)
        except (git.config.cp.NoSectionError, git.config.cp.NoOptionError) as exc:
            raise colrev_exceptions.CoLRevException(
                "Global git variables (user name and email) not available."
            ) from exc
        return global_conf_details

    def check_git_installed(self) -> None:  # pragma: no cover
        """Check whether git is installed"""

        try:
            git_instance = git.Git()
            _ = git_instance.version()
        except git.GitCommandNotFound as exc:
            print(exc)

    def _get_status(self, review_manager: colrev.review_manager.ReviewManager) -> dict:
        status_dict = {}
        status_yml = review_manager.paths.status
        with open(status_yml, encoding="utf8") as stream:
            try:
                status_dict = yaml.safe_load(stream)
            except yaml.YAMLError as exc:  # pragma: no cover
                print(exc)
        return status_dict

    def get_environment_details(self) -> dict:
        """Get the environment details"""

        environment_details = {}
        size = 0
        last_modified = "NOT_INITIATED"
        status = "TODO"

        size = 0
        environment_details["index"] = {
            "size": size,
            "last_modified": last_modified,
            "path": str(Filepaths.LOCAL_ENVIRONMENT_DIR),
            "status": status,
        }

        environment_stats = self._get_environment_stats()

        environment_details["local_repos"] = {
            "repos": environment_stats["repos"],
            "broken_links": environment_stats["broken_links"],
        }
        return environment_details

    def _get_environment_stats(self) -> dict:
        """Get the environment stats"""

        local_repos = self.local_repos()
        repos = []
        broken_links = []
        for repo in local_repos:
            try:
                cp_review_manager = colrev.review_manager.ReviewManager(
                    path_str=repo["repo_source_path"]
                )
                check_operation = colrev.ops.check.CheckOperation(cp_review_manager)
                repo_stat = self._get_status(cp_review_manager)
                repo["size"] = repo_stat["overall"]["md_processed"]
                repo["progress"] = -1
                if repo_stat["atomic_steps"] != 0:
                    repo["progress"] = round(
                        repo_stat["completed_atomic_steps"] / repo_stat["atomic_steps"],
                        2,
                    )

                git_repo = check_operation.review_manager.dataset.get_repo()
                repo["remote"] = bool(git_repo.remotes)
                repo["behind_remote"] = (
                    check_operation.review_manager.dataset.behind_remote()
                )

                repos.append(repo)
            except (
                colrev_exceptions.CoLRevException,
                git.InvalidGitRepositoryError,
            ):  # pragma: no cover
                broken_links.append(repo)
        return {"repos": repos, "broken_links": broken_links}

    def get_curated_outlets(self) -> list:
        """Get the curated outlets"""
        curated_outlets: typing.List[str] = []
        for repo_source_path in [
            x["repo_source_path"]
            for x in self.local_repos()
            if ".colrev/curated_metadata/" in x["repo_source_path"]
        ]:
            try:
                with open(f"{repo_source_path}/readme.md", encoding="utf-8") as file:
                    first_line = file.readline()
                curated_outlets.append(first_line.lstrip("# ").replace("\n", ""))

                with open(
                    f"{repo_source_path}/data/records.bib", encoding="utf-8"
                ) as file:
                    outlets = []
                    for line in file.readlines():
                        # Note : the second part ("journal:"/"booktitle:")
                        # ensures that data provenance fields are skipped
                        if (
                            Fields.JOURNAL == line.lstrip()[:7]
                            and "journal:" != line.lstrip()[:8]
                        ):
                            journal = line[line.find("{") + 1 : line.rfind("}")]
                            if journal != FieldValues.UNKNOWN:
                                outlets.append(journal)
                        if (
                            line.lstrip()[:9] == Fields.BOOKTITLE
                            and line.lstrip()[:10] != "booktitle:"
                        ):
                            booktitle = line[line.find("{") + 1 : line.rfind("}")]
                            if booktitle != FieldValues.UNKNOWN:
                                outlets.append(booktitle)

                    if len(set(outlets)) > 1:  # pragma: no cover
                        raise colrev_exceptions.CuratedOutletNotUnique(
                            "Error: Duplicate outlets in curated_metadata of "
                            f"{repo_source_path} : {','.join(list(set(outlets)))}"
                        )
            except FileNotFoundError as exc:  # pragma: no cover
                print(exc)

        return curated_outlets

    def _dict_keys_exists(self, element: dict, *keys: str) -> bool:
        """Check if *keys (nested) exists in `element` (dict)."""
        if len(keys) < 2:
            raise AttributeError(
                "keys_exists() expects at least two arguments, one given."
            )

        _element = element
        for key in keys:
            try:
                _element = _element[key]
            except KeyError:
                return False
        return True

    def get_settings_by_key(self, key: str) -> str | None:
        """Loads setting by the given key

        The registry is stored in /home/username/colrev/registry.json"""
        environment_registry = self.load_environment_registry()
        keys = key.split(".")
        if self._dict_keys_exists(environment_registry, *keys):
            return get_by_path(environment_registry, keys)
        return None

    def update_registry(self, key: str, value: str) -> None:
        """Updates a given key in the registry with new value

        The registry is stored in /home/username/colrev/registry.json"""

        keys = key.split(".")
        # We don't want to allow user to replace any core settings, so check for packages key
        if keys[0] != "packages" or len(keys) < 2:
            raise colrev_exceptions.PackageSettingMustStartWithPackagesException(key)
        self.environment_registry = self.load_environment_registry()
        dict_set_nested(self.environment_registry, keys, value)
        self.save_environment_registry(self.environment_registry)



#! /usr/bin/env python
"""Convenience functions to load files (BiBTeX, RIS, CSV, etc.)

Usage::

    import colrev.loader.load_utils

    # Files
    records = colrev.loader.load_utils.load(filename=filename, logger=logger)

    # Strings
    records = colrev.loader.load_utils.loads(load_str=load_str, logger=logger)

    returns: records (dict)

Most formats require a mapping from the FIELDS to the standard CoLRev Fields (see CEP 002), which

- can involve merging of FIELDS (e.g. AU / author fields)
- can be conditional upon the ENTRYTYPE (e.g., publication_name: journal or booktitle)

Example BibTeX record::

    @article{Guo2021,
        title    = {How Trust Leads to Commitment on Microsourcing Platforms},
        author   = {Guo, Wenbo and Straub, Detmar W. and Zhang, Pengzhu and Cai, Zhao},
        journal  = {MIS Quarterly},
        year     = {2021}
        volume   = {45},
        number   = {3},
        pages    = {1309--1348},
        url      = {https://aisel.aisnet.org/misq/vol45/iss3/13},
        doi      = {10.25300/MISQ/2021/16100},
    }

Example ENL record::

    %T How Trust Leads to Commitment on Microsourcing Platforms
    %0 Journal Article
    %A Guo, Wenbo
    %A Straub, Detmar W.
    %A Zhang, Pengzhu
    %A Cai, Zhao
    %B Management Information Systems Quarterly
    %D 2021
    %8 September  1, 2021
    %V 45
    %N 3
    %P 1309-1348
    %U https://aisel.aisnet.org/misq/vol45/iss3/13
    %X IS research has extensively examined the role of trust in client-vendor relationships...

Example markdown reference section::

    # References

    Guo, W. and Straub, D. W. and Zhang, P. and Cai, Z. (2021). How Trust Leads to Commitment
          on Microsourcing Platforms. MIS Quarterly, 45(3), 1309--1348.


Example nbib record::

    OWN - ERIC
    TI  - How Trust Leads to Commitment on Microsourcing Platforms
    AU  - Guo, Wenbo
    AU  - Straub, Detmar W.
    AU  - Zhang, Pengzhu
    AU  - Cai, Zhao
    JT  - MIS Quarterly
    DP  - 2021
    VI  - 45
    IP  - 3
    PG  - 1309-1348

Example RIS record::

    TY  - JOUR
    AU  - Guo, Wenbo
    AU  - Straub, Detmar W.
    AU  - Zhang, Pengzhu
    AU  - Cai, Zhao
    DA  - 2021/09/01
    DO  - 10.25300/MISQ/2021/16100
    ID  - Guo2021
    T2  - Management Information Systems Quarterly
    TI  - How Trust Leads to Commitment on Microsourcing Platforms
    VL  - 45
    IS  - 3
    SP  - 1309
    EP  - 1348
    UR  - https://aisel.aisnet.org/misq/vol45/iss3/13
    PB  - Association for Information Systems
    ER  -

Example csv record:

    title;author;year;
    How Trust Leads to Commitment;Guo, W. and Straub, D.;2021;

Example json record:

    [
        {
            "uid": "GS:10955542985646283611",
            "title": "How Trust Leads to Commitment on Microsourcing Platforms: Unraveling the Effects of Governance and Third-Party Mechanisms on Triadic Microsourcing …", # pylint: disable=line-too-long
            "source": "MIS Quarterly",
            "publisher": "search.ebscohost.com",
            "article_url": "https://search.ebscohost.com/login.aspx?direct=true&profile=ehost&scope=site&authtype=crawler&jrnl=02767783&AN=152360579&h=TN%2BrXT9JtZbo1Hx6o6GaYvtoi2ScuJu0iI%2Ba5gvAvmYh9fkUTt3gu0586DSklLEUoLwfC70wD114VB%2BmhsQObw%3D%3D&crl=c", # pylint: disable=line-too-long
            "cites_url": "https://scholar.google.com/scholar?cites=10955542985646283611&as_sdt=2005&sciodt=2007&hl=en", # pylint: disable=line-too-long
            "related_url": "https://scholar.google.com/scholar?q=related:Wxv47EzoCZgJ:scholar.google.com/&scioq=intitle:microsourcing&hl=en&as_sdt=2007", # pylint: disable=line-too-long
            "abstract": "… microsourcing relationship which includes the microsourcer, the microsourcee, and the microsourcing … We argue that the microsourcing phenomenon fits these three critical dimensions …", # pylint: disable=line-too-long
            "rank": 1,
            "year": 2021,
            "volume": 0,
            "issue": 0,
            "startpage": 0,
            "endpage": 0,
            "cites": 9,
            "ecc": 9,
            "use": 1,
            "authors": [
                "W Guo",
                "D Straub",
                "P Zhang",
                "Z Cai"
            ]
        },
    ]

"""
from __future__ import annotations

import logging
import tempfile
import typing
from pathlib import Path

import colrev.loader.bib
import colrev.loader.enl
import colrev.loader.json
import colrev.loader.md
import colrev.loader.nbib
import colrev.loader.ris
import colrev.loader.table


# pylint: disable=too-many-arguments
# flake8: noqa: E501


def load(  # type: ignore
    filename: Path,
    *,
    entrytype_setter: typing.Callable = lambda x: x,
    field_mapper: typing.Callable = lambda x: x,
    id_labeler: typing.Callable = lambda x: x,
    unique_id_field: str = "",
    logger: logging.Logger = logging.getLogger(__name__),
    empty_if_file_not_exists: bool = True,
) -> dict:
    """Load a file and return records as a dictionary"""

    if not filename.is_file():
        if empty_if_file_not_exists:
            return {}
        raise FileNotFoundError

    if filename.suffix == ".bib":
        parser = colrev.loader.bib.BIBLoader  # type: ignore
    elif filename.suffix in [".csv", ".xls", ".xlsx"]:
        parser = colrev.loader.table.TableLoader  # type: ignore
    elif filename.suffix == ".ris":
        parser = colrev.loader.ris.RISLoader  # type: ignore
    elif filename.suffix in [".enl", ".txt"]:
        parser = colrev.loader.enl.ENLLoader  # type: ignore
    elif filename.suffix == ".md":
        parser = colrev.loader.md.MarkdownLoader  # type: ignore
    elif filename.suffix == ".nbib":
        parser = colrev.loader.nbib.NBIBLoader  # type: ignore
    elif filename.suffix == ".json":
        parser = colrev.loader.json.JSONLoader  # type: ignore
    else:
        raise NotImplementedError

    return parser(
        filename=filename,
        entrytype_setter=entrytype_setter,
        field_mapper=field_mapper,
        id_labeler=id_labeler,
        unique_id_field=unique_id_field,
        logger=logger,
    ).load()


def loads(  # type: ignore
    load_string: str,
    *,
    implementation: str,
    entrytype_setter: typing.Callable = lambda x: x,
    field_mapper: typing.Callable = lambda x: x,
    id_labeler: typing.Callable = lambda x: x,
    unique_id_field: str = "",
    logger: logging.Logger = logging.getLogger(__name__),
) -> dict:
    """Load a string and return records as a dictionary"""

    if implementation not in [
        "bib",
        "csv",
        "xls",
        "xlsx",
        "ris",
        "enl",
        "md",
        "nbib",
        "json",
    ]:
        raise NotImplementedError

    with tempfile.NamedTemporaryFile(
        mode="wb", delete=False, suffix=f".{implementation}"
    ) as temp_file:
        temp_file.write(load_string.encode("utf-8"))
        temp_file_path = Path(temp_file.name)

    return load(
        filename=temp_file_path,
        entrytype_setter=entrytype_setter,
        field_mapper=field_mapper,
        id_labeler=id_labeler,
        unique_id_field=unique_id_field,
        logger=logger,
    )


def get_nr_records(  # type: ignore
    filename: Path,
) -> int:
    """Get the number of records in a file"""

    if not filename.exists():
        return 0

    if filename.suffix == ".bib":
        parser = colrev.loader.bib.BIBLoader  # type: ignore
    elif filename.suffix in [".csv", ".xls", ".xlsx"]:
        parser = colrev.loader.table.TableLoader  # type: ignore
    elif filename.suffix == ".ris":
        parser = colrev.loader.ris.RISLoader  # type: ignore
    elif filename.suffix in [".enl", ".txt"]:
        parser = colrev.loader.enl.ENLLoader  # type: ignore
    elif filename.suffix == ".md":
        parser = colrev.loader.md.MarkdownLoader  # type: ignore
    elif filename.suffix == ".nbib":
        parser = colrev.loader.nbib.NBIBLoader  # type: ignore
    elif filename.suffix == ".json":
        parser = colrev.loader.json.JSONLoader  # type: ignore
    else:
        raise NotImplementedError

    return parser.get_nr_records(filename)



#!/usr/bin/env python3
"""Hook to update non-CoLRev repositories"""
from pathlib import Path

import colrev.packages.sync.src.sync
from colrev.constants import Colors
from colrev.constants import ExitCodes
from colrev.constants import Fields


def main() -> int:
    """Main entrypoint for the update hoooks"""

    sync_operation = colrev.packages.sync.src.sync.Sync()
    sync_operation.get_cited_papers()

    if len(sync_operation.non_unique_for_import) > 0:
        # Note: interactive resolution not supported in pre-commit hooks
        print(
            f"To resolve non-unique cases, run {Colors.ORANGE}colrev sync{Colors.END}"
        )

        return ExitCodes.FAIL

    # Add IDs to .spelling (if not already in the file)
    if Path(".spelling").is_file():
        if not all(
            x.data[Fields.ID] in Path(".spelling").read_text(encoding="utf-8")
            for x in sync_operation.records_to_import
        ):
            with open(".spelling", "a", encoding="utf-8") as file:
                for item in sync_operation.records_to_import:
                    file.write(item.data[Fields.ID])

    sync_operation.add_to_bib()

    return ExitCodes.SUCCESS


if __name__ == "__main__":
    raise SystemExit(main())



#! /usr/bin/env python
"""CoLRev dedupe operation: identify and merge duplicate records."""
from __future__ import annotations

import string
import typing
from collections import defaultdict
from itertools import combinations
from pathlib import Path

import pandas as pd
from bib_dedupe.bib_dedupe import prep

import colrev.exceptions as colrev_exceptions
import colrev.process.operation
import colrev.record.record
from colrev.constants import Colors
from colrev.constants import EndpointType
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields
from colrev.constants import OperationsType
from colrev.constants import RecordState


class Dedupe(colrev.process.operation.Operation):
    """Deduplicate records (entity resolution)"""

    NON_DUPLICATE_FILE_XLSX = Path("non_duplicates_to_validate.xlsx")
    NON_DUPLICATE_FILE_TXT = Path("non_duplicates_to_validate.txt")
    DUPLICATES_TO_VALIDATE = Path("duplicates_to_validate.xlsx")
    SAME_SOURCE_MERGE_FILE = Path("same_source_merges.txt")
    PREVENTED_SAME_SOURCE_MERGE_FILE = Path("prevented_same_source_merges.txt")

    debug = False
    type = OperationsType.dedupe

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        notify_state_transition_operation: bool = True,
    ) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=self.type,
            notify_state_transition_operation=notify_state_transition_operation,
        )
        self.dedupe_dir = self.review_manager.paths.dedupe
        self.non_dupe_file_xlsx = self.dedupe_dir / self.NON_DUPLICATE_FILE_XLSX
        self.non_dupe_file_txt = self.dedupe_dir / self.NON_DUPLICATE_FILE_TXT
        self.dupe_file = self.dedupe_dir / self.DUPLICATES_TO_VALIDATE

        self.same_source_merge_file = self.dedupe_dir / self.SAME_SOURCE_MERGE_FILE
        self.prevented_same_source_merge_file = (
            self.dedupe_dir / self.PREVENTED_SAME_SOURCE_MERGE_FILE
        )
        self.dedupe_dir.mkdir(exist_ok=True, parents=True)

    @classmethod
    def _dfs(cls, node: str, graph: dict, visited: dict, component: list) -> None:
        visited[node] = True
        component.append(node)
        for neighbor in graph[node]:
            if not visited[neighbor]:
                cls._dfs(neighbor, graph, visited, component)

    @classmethod
    def connected_components(cls, id_sets: list) -> list:
        """
        Find the connected components in a graph.

        Args:
            id_sets (list): A list of id sets.

        Returns:
            list: A list of connected components.
        """
        graph = defaultdict(list)

        # Create an adjacency list
        for origin_set in id_sets:
            for combination in combinations(origin_set, 2):
                graph[combination[0]].append(combination[1])
                graph[combination[1]].append(combination[0])

        visited = {node: False for node in graph}
        components = []

        for node in graph:
            if not visited[node]:
                component: typing.List[str] = []
                cls._dfs(node, graph, visited, component)
                components.append(sorted(component))

        return components

    @classmethod
    def get_records_for_dedupe(
        cls, *, records_df: pd.DataFrame, verbosity_level: int = 0
    ) -> pd.DataFrame:
        """Get (pre-processed) records for dedupe"""
        return prep(records_df=records_df, verbosity_level=verbosity_level)

    def _select_primary_merge_record(self, rec_1: dict, rec_2: dict) -> list:
        # pylint: disable=too-many-branches

        # Heuristic

        # 1. if both records are prepared (or the same status),
        # merge into the record with the "lower" colrev_id
        if rec_1[Fields.STATUS] == rec_2[Fields.STATUS]:
            if rec_1[Fields.ID][-1].isdigit() and not rec_2[Fields.ID][-1].isdigit():
                main_record = rec_1
                dupe_record = rec_2
            elif (
                not rec_1[Fields.ID][-1].isdigit()
                and not rec_2[Fields.ID][-1].isdigit()
            ):
                # Last characters are letters in both records
                # Select the one that's first in the alphabet
                try:
                    pos_rec_1_suffix = string.ascii_lowercase.index(
                        rec_1[Fields.ID][-1]
                    )
                except ValueError:
                    pos_rec_1_suffix = -1
                try:
                    pos_rec_2_suffix = string.ascii_lowercase.index(
                        rec_2[Fields.ID][-1]
                    )
                except ValueError:
                    pos_rec_2_suffix = -1

                if pos_rec_1_suffix < pos_rec_2_suffix:
                    main_record = rec_1
                    dupe_record = rec_2
                else:
                    dupe_record = rec_2
                    main_record = rec_1
            else:
                main_record = rec_2
                dupe_record = rec_1

        # 2. If a record is md_prepared, use it as the dupe record
        elif rec_1[Fields.STATUS] == RecordState.md_prepared:
            main_record = rec_2
            dupe_record = rec_1
        elif rec_2[Fields.STATUS] == RecordState.md_prepared:
            main_record = rec_1
            dupe_record = rec_2

        # 3. If a record is md_processed, use the other record as the dupe record
        # -> during the fix_errors procedure, records are in md_processed
        # and beyond.
        elif rec_1[Fields.STATUS] == RecordState.md_processed:
            main_record = rec_1
            dupe_record = rec_2
        elif rec_2[Fields.STATUS] == RecordState.md_processed:
            main_record = rec_2
            dupe_record = rec_1

        # 4. Merge into curated record (otherwise)
        else:
            if colrev.record.record.Record(rec_2).masterdata_is_curated():
                main_record = rec_2
                dupe_record = rec_1
            else:
                main_record = rec_1
                dupe_record = rec_2

        return [main_record, dupe_record]

    def _same_source_merge(
        self,
        *,
        main_record: colrev.record.record.Record,
        dupe_record: colrev.record.record.Record,
    ) -> bool:
        main_rec_sources = [x.split("/")[0] for x in main_record.data[Fields.ORIGIN]]
        dupe_rec_sources = [x.split("/")[0] for x in dupe_record.data[Fields.ORIGIN]]
        same_sources = set(main_rec_sources).intersection(set(dupe_rec_sources))

        if len(same_sources) == 0:
            return False

        # don't raise unproblematic same-source merges
        # if same_sources start with md_... AND have same IDs: no same-source merge.
        if all(x.startswith("md_") for x in same_sources):
            main_rec_same_source_origins = [
                x
                for x in main_record.data[Fields.ORIGIN]
                if x.split("/")[0] in same_sources
            ]
            dupe_rec_same_source_origins = [
                x
                for x in dupe_record.data[Fields.ORIGIN]
                if x.split("/")[0] in same_sources
            ]
            if set(main_rec_same_source_origins) == set(dupe_rec_same_source_origins):
                return False
            return True
        return True

    def _notify_on_merge(
        self,
        *,
        main_record: colrev.record.record.Record,
        dupe_record: colrev.record.record.Record,
    ) -> None:
        merge_info = main_record.data[Fields.ID] + " - " + dupe_record.data[Fields.ID]
        if not self._same_source_merge(
            main_record=main_record, dupe_record=dupe_record
        ):
            self.review_manager.logger.info(
                f" merge {main_record.data[Fields.ID]} - {dupe_record.data[Fields.ID]}"
            )
            return

        self.review_manager.logger.warning(
            f"{Colors.ORANGE}"
            "Apply same source merge "
            f"{Colors.END} "
            f"{merge_info}\n"
            f"  {main_record.format_bib_style()}\n"
            f"  {dupe_record.format_bib_style()}\n"
            f"  {main_record.data.get(Fields.ORIGIN, ['ERROR'])} x "
            f"{dupe_record.data.get(Fields.ORIGIN, ['ERROR'])}\n"
        )

    def _is_cross_level_merge(
        self,
        *,
        main_record: colrev.record.record.Record,
        dupe_record: colrev.record.record.Record,
    ) -> bool:
        is_cross_level_merge_attempt = False
        if main_record.data[Fields.ENTRYTYPE] in [
            ENTRYTYPES.PROCEEDINGS
        ] or dupe_record.data[Fields.ENTRYTYPE] in [ENTRYTYPES.PROCEEDINGS]:
            is_cross_level_merge_attempt = True

        if (
            main_record.data[Fields.ENTRYTYPE] == ENTRYTYPES.BOOK
            and dupe_record.data[Fields.ENTRYTYPE] == ENTRYTYPES.INBOOK
        ):
            is_cross_level_merge_attempt = True

        if (
            main_record.data[Fields.ENTRYTYPE] == ENTRYTYPES.INBOOK
            and dupe_record.data[Fields.ENTRYTYPE] == ENTRYTYPES.BOOK
        ):
            is_cross_level_merge_attempt = True

        return is_cross_level_merge_attempt

    # pylint: disable=too-many-arguments
    def _print_merge_stats(
        self,
        *,
        records: dict,
        duplicate_id_mappings: dict,
        removed_duplicates: list,
        complete_dedupe: bool,
        set_to_md_processed: list,
    ) -> None:
        print()
        self.review_manager.logger.info("Summary:")
        for record_id, duplicate_ids in duplicate_id_mappings.items():
            duplicate_ids = [x.replace(record_id, "-") for x in duplicate_ids]
            if record_id not in records:
                continue
            if RecordState.md_prepared == records[record_id][Fields.STATUS]:
                if complete_dedupe:
                    self.review_manager.logger.info(
                        f" {Colors.GREEN}{record_id} ({'|'.join(duplicate_ids)}) ".ljust(
                            46
                        )
                        + f"md_prepared →  md_processed{Colors.END}"
                    )
                else:
                    self.review_manager.logger.info(
                        f" {Colors.GREEN}{record_id} ({'|'.join(duplicate_ids)}) ".ljust(
                            46
                        )
                        + f"md_prepared →  md_prepared{Colors.END}"
                    )
        if complete_dedupe:
            for rid in set_to_md_processed:
                self.review_manager.logger.info(
                    f" {Colors.GREEN}{rid}".ljust(46)
                    + f"md_prepared →  md_processed{Colors.END}"
                )

        self.review_manager.logger.info(
            "Merged duplicates: ".ljust(39) + f"{len(removed_duplicates)} records"
        )

    def _apply_records_merges(
        self, *, records: dict, removed_duplicates: list, complete_dedupe: bool
    ) -> list:
        for record in records.values():
            if "MOVED_DUPE_ID" in record:
                del record["MOVED_DUPE_ID"]
        for removed_duplicate in removed_duplicates:
            if removed_duplicate in records:
                del records[removed_duplicate]

        set_to_md_processed = []
        if complete_dedupe:
            # Set remaining records to md_processed (not duplicate) because all records
            # have been considered by dedupe
            for record_dict in records.values():
                if record_dict[Fields.STATUS] == RecordState.md_prepared:
                    record = colrev.record.record.Record(record_dict)
                    record.set_status(RecordState.md_processed)
                    set_to_md_processed.append(record.data[Fields.ID])

        self.review_manager.dataset.save_records_dict(records)
        return set_to_md_processed

    def _skip_merge_condition(
        self,
        *,
        main_record: colrev.record.record.Record,
        dupe_record: colrev.record.record.Record,
    ) -> bool:
        if self.review_manager.force_mode:
            return False
        if self._is_cross_level_merge(main_record=main_record, dupe_record=dupe_record):
            self.review_manager.logger.info(
                "Prevented cross-level merge: "
                f"{main_record.data[Fields.ID]} - {dupe_record.data[Fields.ID]}"
            )
            return True

        # if (
        #     self._same_source_merge(main_record=main_record, dupe_record=dupe_record)
        #     and self.policy == colrev.settings.SameSourceMergePolicy.prevent
        # ):
        #     self.review_manager.logger.info(
        #         "Prevented same-source merge: "
        #         f"{main_record.data[Fields.ID]} - {dupe_record.data[Fields.ID]}"
        #     )
        #     return True

        return False

    def _update_duplicate_id_mappings(
        self,
        *,
        duplicate_id_mappings: dict,
        main_record: colrev.record.record.Record,
        dupe_record: colrev.record.record.Record,
    ) -> None:
        if main_record.data[Fields.ID] not in duplicate_id_mappings:
            duplicate_id_mappings[main_record.data[Fields.ID]] = [
                dupe_record.data[Fields.ID]
            ]
        else:
            duplicate_id_mappings[main_record.data[Fields.ID]].append(
                dupe_record.data[Fields.ID]
            )

    def apply_merges(
        self,
        *,
        id_sets: list,
        complete_dedupe: bool = False,
        preferred_masterdata_sources: typing.Optional[list] = None,
    ) -> None:
        """Apply deduplication decisions

        id_sets : [[ID_1, ID_2, ID_3], ...]

        - complete_dedupe: when not all potential duplicates were considered,
        we cannot set records to md_procssed for non-duplicate decisions
        """

        preferred_masterdata_source_prefixes = []
        if preferred_masterdata_sources:
            preferred_masterdata_source_prefixes = [
                s.get_origin_prefix() for s in preferred_masterdata_sources
            ]

        records = self.review_manager.dataset.load_records_dict()
        non_existing_ids = [
            ID
            for ID in [id for id_set in id_sets for id in id_set]
            if ID not in records
        ]
        if non_existing_ids:
            print(f"Non-existing IDs: {non_existing_ids}")
            print(f"Records IDs: {records.keys()}")
        assert not non_existing_ids, "Not all IDs from id_sets are present in records"

        # Notify users about items with only one unique ID
        for id_set in id_sets:
            if len(set(id_set)) == 1:
                self.review_manager.logger.info(
                    f"Skipping merge for identical IDs: {id_set[0]}"
                )
        # Drop cases where IDs are identical
        id_sets = [id_set for id_set in id_sets if len(set(id_set)) != 1]

        removed_duplicates = []
        duplicate_id_mappings: typing.Dict[str, list] = {}
        for main_record, dupe_record in self._get_records_to_merge(
            records=records, id_sets=id_sets
        ):
            if self._skip_merge_condition(
                main_record=main_record, dupe_record=dupe_record
            ):
                continue

            self._notify_on_merge(
                main_record=main_record,
                dupe_record=dupe_record,
            )

            self._update_duplicate_id_mappings(
                duplicate_id_mappings=duplicate_id_mappings,
                main_record=main_record,
                dupe_record=dupe_record,
            )
            dupe_record.data["MOVED_DUPE_ID"] = main_record.data[Fields.ID]
            main_record.merge(
                dupe_record,
                default_source="merged",
                preferred_masterdata_source_prefixes=preferred_masterdata_source_prefixes,
            )
            removed_duplicates.append(dupe_record.data[Fields.ID])

        set_to_md_processed = self._apply_records_merges(
            records=records,
            removed_duplicates=removed_duplicates,
            complete_dedupe=complete_dedupe,
        )

        self._print_merge_stats(
            records=records,
            duplicate_id_mappings=duplicate_id_mappings,
            removed_duplicates=removed_duplicates,
            complete_dedupe=complete_dedupe,
            set_to_md_processed=set_to_md_processed,
        )

    def _get_records_to_merge(
        self, *, records: dict, id_sets: list
    ) -> typing.Iterable[tuple]:
        """Resolves multiple/chained duplicates (by following the MOVED_DUPE_ID mark)
        and returns tuples with the primary merge record in the first position."""

        for id_set in id_sets:
            recs_to_merge = [r for r in records.values() if r[Fields.ID] in id_set]

            # Simple way of implementing the closure
            # cases where the main_record has already been merged into another record
            for ind, rec_to_merge in enumerate(recs_to_merge):
                cur_rec = rec_to_merge
                while "MOVED_DUPE_ID" in cur_rec:
                    cur_rec = records[cur_rec["MOVED_DUPE_ID"]]
                recs_to_merge[ind] = cur_rec

            if all(recs_to_merge[0][Fields.ID] == r[Fields.ID] for r in recs_to_merge):
                continue

            for rec_1, rec_2 in combinations(recs_to_merge, 2):
                main_record_dict, dupe_record_dict = self._select_primary_merge_record(
                    rec_1, rec_2
                )

                main_record = colrev.record.record.Record(main_record_dict)
                dupe_record = colrev.record.record.Record(dupe_record_dict)

                yield (main_record, dupe_record)

    def _get_origins_for_current_ids(self, current_record_ids: list) -> dict:
        """
        For each record ID, get the origins from the most recent history entry.
        """

        ids_origins: typing.Dict[str, typing.List[str]] = {
            rid: [] for rid in current_record_ids
        }
        # history = next(self.review_manager.dataset.load_records_from_history(), None)
        # if history:
        #     for rid in ids_origins:
        #         ids_origins[rid] = history.get(rid, {}).get(Fields.ORIGIN, [])

        records = self.review_manager.dataset.load_records_dict()
        for rid in ids_origins:
            ids_origins[rid] = records.get(rid, {}).get(Fields.ORIGIN, [])
            if ids_origins[rid] == []:
                raise colrev_exceptions.RecordNotFoundException(
                    f"Did not find record with ID {rid} for unmerge"
                )

        return ids_origins

    def _remove_merged_records(self, records: dict, ids_origins: dict) -> None:
        """
        Remove records that have been merged from the records dictionary.
        These records are identified by their current IDs.
        """
        for rid in ids_origins:
            records.pop(rid, None)

    def _revert_merge_for_records(
        self, unmerged_records: dict, ids_origins: dict
    ) -> dict:
        """
        Attempt to revert the merge operation for each record based on its origins.
        """

        for hist_recs in self.review_manager.dataset.load_records_from_history():
            for rid in list(ids_origins.keys()):
                origins = ids_origins[rid]

                unmerged_rids = []

                for hist_rec in hist_recs.values():
                    if not set(hist_rec.get(Fields.ORIGIN, [])).intersection(
                        set(origins)
                    ):
                        continue

                    # skip if hist_recs still contains the merged records (identical origin set)
                    # ie., need to consider older commits
                    if origins == hist_recs[rid].get(Fields.ORIGIN, []):
                        continue
                    if any(orig in origins for orig in hist_rec.get(Fields.ORIGIN, [])):
                        assert hist_rec[Fields.ID] not in unmerged_records
                        hist_rec.update({Fields.STATUS: RecordState.md_processed})
                        self.review_manager.logger.info(
                            f"add historical record: {hist_rec[Fields.ID]}"
                        )
                        unmerged_records[hist_rec[Fields.ID]] = hist_rec
                        unmerged_rids.append(rid)

                for unmerged_rid in set(unmerged_rids):
                    ids_origins.pop(unmerged_rid)

            # Stop if all unmerged records were restored
            if not ids_origins:
                break

        return unmerged_records

    def unmerge_records(
        self,
        *,
        current_record_ids: list,
    ) -> None:
        """Unmerge duplicate decision of the records, as identified by their ids.

        The current_record_ids identifies the records by their current IDs and
        unmerges their most recent merge in history.

        """
        # Map each current record ID to its origins for easy lookup of historical records.
        ids_origins = self._get_origins_for_current_ids(current_record_ids)
        print(ids_origins)

        # Load the current state of all records.
        records = self.review_manager.dataset.load_records_dict()

        # Remove the merged records from the current state.
        self._remove_merged_records(records, ids_origins)
        print(f"After removal: {records.keys()}")

        # Attempt to revert the merge operation for each record.
        records = self._revert_merge_for_records(records, ids_origins)
        print(f"After revert: {records.keys()}")

        self.review_manager.dataset.save_records_dict(records)

    def fix_errors(self, *, false_positives: list, false_negatives: list) -> None:
        """Fix lists of errors"""

        self.unmerge_records(current_record_ids=false_positives)
        self.apply_merges(id_sets=false_negatives, complete_dedupe=False)

        if self.review_manager.dataset.records_changed():
            self.review_manager.dataset.create_commit(
                msg="Validate and correct duplicates",
                manual_author=True,
            )

    def get_info(self) -> dict:
        """Get info on cuts (overlap of search sources) and same source merges"""

        # same_source_merges: typing.Any = []
        # source_overlaps: typing.Any = []
        # info = {
        #     "same_source_merges": same_source_merges,
        #     "source_overlaps": source_overlaps,
        # }
        # return info
        raise NotImplementedError

    def merge_records(self, *, merge: list) -> None:
        """Merge records by ID sets"""

        self.apply_merges(id_sets=merge)

    def merge_based_on_global_ids(self, *, apply: bool = False) -> None:
        """Merge records based on global IDs (e.g., doi)"""

        # pylint: disable=too-many-branches

        self.review_manager.logger.info(
            "Dedupe operation [merge based on identical global_ids]"
        )

        records = self.review_manager.dataset.load_records_dict()

        global_keys = [
            Fields.DOI,
            Fields.COLREV_ID,
            Fields.PUBMED_ID,
            Fields.DBLP_KEY,
            Fields.URL,
        ]
        if Fields.COLREV_ID in global_keys:
            for record in records.values():
                try:
                    record[Fields.COLREV_ID] = colrev.record.record.Record(
                        data=record
                    ).get_colrev_id()
                except colrev_exceptions.NotEnoughDataToIdentifyException:
                    pass

        id_sets = []
        for global_key in global_keys:
            global_key_dict: typing.Dict[str, list] = {}
            for record in records.values():
                if global_key not in record:
                    continue
                if record[global_key] in global_key_dict:
                    global_key_dict[record[global_key]].append(record[Fields.ID])
                else:
                    global_key_dict[record[global_key]] = [record[Fields.ID]]

            global_key_duplicates = [v for v in global_key_dict.values() if len(v) > 1]
            for duplicate in global_key_duplicates:
                id_sets.extend(list(combinations(duplicate, 2)))

        if apply:
            self.apply_merges(id_sets=id_sets)
            self.review_manager.dataset.create_commit(
                msg="Merge records (identical global IDs)"
            )

        elif id_sets:
            print()
            self.review_manager.logger.info("Found records with identical global-ids:")
            self.review_manager.logger.info(
                f"{Colors.ORANGE}To merge records with identical global-ids, "
                f"run colrev merge -gid{Colors.END}"
            )
            print()

        else:
            print()

    @colrev.process.operation.Operation.decorate()
    def main(self, *, debug: bool = False) -> None:
        """Dedupe records (main entrypoint)"""

        self.review_manager.logger.info("Dedupe")
        self.review_manager.logger.info(
            "Identifies duplicate records and merges them (keeping traces to their origins)."
        )
        self.review_manager.logger.info(
            "See https://colrev-environment.github.io/colrev/manual/metadata_retrieval/dedupe.html"
        )
        self.debug = debug

        if not self.review_manager.high_level_operation:
            print()

        package_manager = self.review_manager.get_package_manager()
        for (
            dedupe_package_endpoint
        ) in self.review_manager.settings.dedupe.dedupe_package_endpoints:
            # Note : load package/script at this point because the same script
            # may run with different parameters

            dedupe_class = package_manager.get_package_endpoint_class(
                package_type=EndpointType.dedupe,
                package_identifier=dedupe_package_endpoint["endpoint"],
            )
            endpoint = dedupe_class(
                dedupe_operation=self, settings=dedupe_package_endpoint
            )

            endpoint.run_dedupe()  # type: ignore
            if not self.review_manager.high_level_operation:
                print()

        dedupe_commit_id = self.review_manager.dataset.get_repo().head.commit.hexsha
        self.review_manager.logger.info("To validate the changes, use")

        self.review_manager.logger.info(
            f"{Colors.ORANGE}colrev validate {dedupe_commit_id}{Colors.END}"
        )
        print()

        self.review_manager.logger.info(
            f"{Colors.GREEN}Completed dedupe operation{Colors.END}"
        )

        if not self.review_manager.settings.prescreen.prescreen_package_endpoints:
            self.review_manager.logger.info("Skipping prescreen/including all records")
            records = self.review_manager.dataset.load_records_dict()
            for record_dict in records.values():
                record = colrev.record.record.Record(record_dict)
                if RecordState.md_processed == record.data[Fields.STATUS]:
                    record.set_status(RecordState.rev_prescreen_included)

            self.review_manager.dataset.save_records_dict(records)
            self.review_manager.dataset.create_commit(msg="Skip prescreen/include all")



#! /usr/bin/env python
"""Checker for isbn-not-matching-pattern."""
from __future__ import annotations

import re
import typing

from colrev.constants import DefectCodes
from colrev.constants import Fields

if typing.TYPE_CHECKING:  # pragma: no cover
    import colrev.record.qm.quality_model

# pylint: disable=too-few-public-methods


class ISBNPatternChecker:
    """The ISBNPatternChecker"""

    msg = DefectCodes.ISBN_NOT_MATCHING_PATTERN

    _ISBN_REGEX = re.compile(
        "^(?:ISBN(?:-1[03])?:? )?(?=[-0-9 ]{17}$|[-0-9X ]{13}$|[0-9X]{10}$)|"
        "(?:97[89][- ]?)?[0-9]{1,5}[- ]?(?:[0-9]+[- ]?){2}[0-9X]$"
    )

    def __init__(
        self, quality_model: colrev.record.qm.quality_model.QualityModel
    ) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.record.Record) -> None:
        """Run the isbn-not-matching-pattern checks"""

        if Fields.ISBN not in record.data or record.ignored_defect(
            key=Fields.ISBN, defect=self.msg
        ):
            return

        if not re.match(self._ISBN_REGEX, record.data[Fields.ISBN]):
            record.add_field_provenance_note(key=Fields.ISBN, note=self.msg)
        else:
            record.remove_field_provenance_note(key=Fields.ISBN, note=self.msg)


def register(quality_model: colrev.record.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(ISBNPatternChecker(quality_model))



#! /usr/bin/env python
"""Export of bib/pdfs as a prep-man operation"""
from __future__ import annotations

import platform
import typing
from pathlib import Path

import pandas as pd
import pymupdf
import zope.interface
from pydantic import BaseModel
from pydantic import Field

import colrev.env.utils
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Colors
from colrev.constants import DefectCodes
from colrev.constants import Fields
from colrev.constants import FieldValues
from colrev.constants import RecordState
from colrev.writer.write_utils import write_file

# pylint: disable=duplicate-code
# pylint: disable=too-few-public-methods
# pylint: disable=too-many-instance-attributes


class ExportManPrepSettings(
    colrev.package_manager.package_settings.DefaultSettings, BaseModel
):
    """Settings for ExportManPrep"""

    endpoint: str
    pdf_handling_mode: str = "symlink"

    _details = {
        "pdf_handling_mode": {
            "tooltip": "Indicates how linked PDFs are handled (symlink/copy_first_page)"
        },
    }


@zope.interface.implementer(colrev.package_manager.interfaces.PrepManInterface)
class ExportManPrep:
    """Manual preparation based on exported and imported metadata (and PDFs if any)"""

    settings: ExportManPrepSettings
    ci_supported: bool = Field(default=False)

    RELATIVE_PREP_MAN_PATH = Path("records_prep_man.bib")
    RELATIVE_PREP_MAN_INFO_PATH = Path("records_prep_man_info.csv")
    RELATIVE_PREP_MAN_INFO_PATH_XLS = Path("records_prep_man_info.xlsx")

    _FIELDS_TO_KEEP = [
        Fields.ENTRYTYPE,
        Fields.AUTHOR,
        Fields.TITLE,
        Fields.YEAR,
        Fields.JOURNAL,
        Fields.BOOKTITLE,
        Fields.STATUS,
        Fields.VOLUME,
        Fields.NUMBER,
        Fields.PAGES,
        Fields.DOI,
        Fields.FILE,
    ]

    settings_class = ExportManPrepSettings

    def __init__(
        self,
        *,
        prep_man_operation: colrev.ops.prep_man.PrepMan,
        settings: dict,
    ) -> None:
        if "pdf_handling_mode" not in settings:
            settings["pdf_handling_mode"] = "symlink"
        assert settings["pdf_handling_mode"] in ["symlink", "copy_first_page"]

        self.settings = self.settings_class(**settings)

        self.review_manager = prep_man_operation.review_manager
        self.quality_model = self.review_manager.get_qm()
        self.prep_dir = self.review_manager.paths.prep
        self.prep_dir.mkdir(exist_ok=True, parents=True)
        self.prep_man_bib_path = self.prep_dir / self.RELATIVE_PREP_MAN_PATH
        self.prep_man_csv_path = self.prep_dir / self.RELATIVE_PREP_MAN_INFO_PATH
        self.prep_man_xlsx_path = self.prep_dir / self.RELATIVE_PREP_MAN_INFO_PATH_XLS

    def _copy_files_for_man_prep(self, *, records: dict) -> None:
        prep_man_path_pdfs = self.prep_dir / Path("pdfs")
        if prep_man_path_pdfs.is_dir():
            input(f"Remove {prep_man_path_pdfs} and press Enter.")
        prep_man_path_pdfs.mkdir(exist_ok=True, parents=True)

        for record in records.values():
            if Fields.FILE in record:
                target_path = self.prep_dir / Path(record[Fields.FILE])
                target_path.parents[0].mkdir(exist_ok=True, parents=True)

                if self.settings.pdf_handling_mode == "symlink":
                    try:
                        target_path.symlink_to(Path(record[Fields.FILE]).resolve())
                    except FileExistsError:
                        pass

                if self.settings.pdf_handling_mode == "copy_first_page":
                    doc1 = pymupdf.Document(str(record[Fields.FILE]))
                    if doc1.page_count > 0:
                        doc2 = pymupdf.Document()
                        doc2.insert_pdf(doc1, to_page=0)
                        doc2.save(str(target_path))

    def _export_prep_man(
        self,
        *,
        records: typing.Dict[str, typing.Dict],
    ) -> None:
        man_prep_recs = {
            k: v
            for k, v in records.items()
            if RecordState.md_needs_manual_preparation == v[Fields.STATUS]
        }

        # Filter out fields that are not needed for manual preparation
        filtered_man_prep_recs = {}
        for citation, fields in man_prep_recs.copy().items():
            for key in fields.copy():
                if key not in self._FIELDS_TO_KEEP:
                    del fields[key]
            filtered_man_prep_recs.update({citation: fields})

        write_file(records_dict=filtered_man_prep_recs, filename=self.prep_man_bib_path)

        if any(Fields.FILE in r for r in man_prep_recs.values()):
            self._copy_files_for_man_prep(records=man_prep_recs)

    def _create_info_dataframe(
        self,
        *,
        records: typing.Dict[str, typing.Dict],
    ) -> None:
        man_prep_recs = [
            v
            for _, v in records.items()
            if RecordState.md_needs_manual_preparation == v[Fields.STATUS]
        ]

        man_prep_info = []
        for record in man_prep_recs:
            for field, value in record[Fields.MD_PROV].items():
                if value["note"] and value["note"] != f"IGNORE:{DefectCodes.MISSING}":
                    man_prep_info.append(
                        {
                            Fields.ID: record[Fields.ID],
                            "field": field,
                            "note": value["note"],
                        }
                    )

        man_prep_info_df = pd.DataFrame(man_prep_info)
        if platform.system() == "Windows":
            # until https://github.com/pylint-dev/pylint/issues/3060 is resolved
            # pylint: disable=abstract-class-instantiated
            with pd.ExcelWriter(self.prep_man_xlsx_path) as writer:
                man_prep_info_df.to_excel(writer, index=False)
        else:
            man_prep_info_df.to_csv(self.prep_man_csv_path, index=False)

    def _drop_unnecessary_provenance_fiels(
        self, *, record: colrev.record.record.Record
    ) -> None:
        colrev_data_provenance_keys_to_drop = []
        for key, items in record.data.get(Fields.D_PROV, {}).items():
            if (
                key not in record.data
                and f"IGNORE:{DefectCodes.MISSING}" not in items["note"]
            ):
                colrev_data_provenance_keys_to_drop.append(key)
        for colrev_data_provenance_key_to_drop in colrev_data_provenance_keys_to_drop:
            del record.data[Fields.D_PROV][colrev_data_provenance_key_to_drop]

        md_prov_keys_to_drop = []
        for key, items in record.data.get(Fields.MD_PROV, {}).items():
            if (
                key not in record.data
                and f"IGNORE:{DefectCodes.MISSING}" not in items["note"]
            ):
                md_prov_keys_to_drop.append(key)
        for md_prov_key_to_drop in md_prov_keys_to_drop:
            del record.data[Fields.MD_PROV][md_prov_key_to_drop]

    def _update_original_record_based_on_man_prepped(
        self,
        *,
        original_record: colrev.record.record.Record,
        man_prepped_record_dict: dict,
    ) -> None:
        dropped_keys = [
            k
            for k in original_record.data
            if (k not in man_prepped_record_dict) and k in self._FIELDS_TO_KEEP
        ]

        if (
            original_record.data[Fields.ENTRYTYPE]
            != man_prepped_record_dict[Fields.ENTRYTYPE]
        ):
            original_record.change_entrytype(
                new_entrytype=man_prepped_record_dict[Fields.ENTRYTYPE],
                qm=self.quality_model,
            )

        if (
            man_prepped_record_dict[Fields.ENTRYTYPE]
            != original_record.data[Fields.ENTRYTYPE]
        ):
            original_record.data[Fields.ENTRYTYPE] = man_prepped_record_dict[
                Fields.ENTRYTYPE
            ]
            original_record.run_quality_model(self.quality_model)

        for key, value in man_prepped_record_dict.items():
            if key in [Fields.STATUS]:
                continue
            if (
                value != original_record.data.get(key, "")
                and value != FieldValues.UNKNOWN
            ):
                original_record.update_field(
                    key=key, value=value, source="man_prep", append_edit=False
                )
                original_record.remove_field_provenance_note(key=key, note="missing")

        for dropped_key in dropped_keys:
            original_record.remove_field(
                key=dropped_key, not_missing_note=True, source="man_prep"
            )

    def _print_stats(self, *, original_record: colrev.record.record.Record) -> None:
        if original_record.data[Fields.STATUS] == RecordState.rev_prescreen_excluded:
            self.review_manager.logger.info(
                f" {Colors.RED}{original_record.data['ID']}".ljust(46)
                + "md_needs_manual_preparation →  rev_prescreen_excluded"
                + f"{Colors.END}"
            )
        elif original_record.data[Fields.STATUS] == RecordState.md_prepared:
            self.review_manager.logger.info(
                f" {Colors.GREEN}{original_record.data['ID']}".ljust(46)
                + "md_needs_manual_preparation →  md_prepared"
                + f"{Colors.END}"
            )
        else:
            man_prep_note = ", ".join(
                k + ":" + v["note"]
                for k, v in original_record.data[Fields.MD_PROV].items()
                if v["note"] != ""
            )
            self.review_manager.logger.info(
                f" {Colors.ORANGE}{original_record.data['ID']}".ljust(46)
                + f"{man_prep_note}"
                + f"{Colors.END}"
            )

    def _import_record(
        self,
        *,
        man_prepped_record_dict: dict,
        original_record: colrev.record.record.Record,
        imported_records: list,
    ) -> None:
        imported_records.append(original_record.data[Fields.ID])
        override = man_prepped_record_dict[Fields.STATUS] == RecordState.md_prepared

        self._update_original_record_based_on_man_prepped(
            original_record=original_record,
            man_prepped_record_dict=man_prepped_record_dict,
        )

        self._drop_unnecessary_provenance_fiels(record=original_record)
        if man_prepped_record_dict[Fields.STATUS] == RecordState.rev_prescreen_excluded:
            original_record.set_status(
                target_state=RecordState.rev_prescreen_excluded,
                force=True,
            )

        else:
            original_record.run_quality_model(self.quality_model, set_prepared=True)

        if override:
            original_record.set_status(RecordState.md_prepared, force=True)

        self._print_stats(original_record=original_record)

    def _import_prep_man(self) -> None:
        self.review_manager.logger.info(
            "Load import changes from "
            f"{self.prep_man_bib_path.relative_to(self.review_manager.path)}"
        )

        man_prep_recs = colrev.loader.load_utils.load(
            filename=self.prep_man_bib_path,
            logger=self.review_manager.logger,
        )

        imported_records: typing.List[dict] = []
        records = self.review_manager.dataset.load_records_dict()
        for record_id, record_dict in records.items():
            if (
                record_dict[Fields.STATUS]
                == RecordState.rev_prescreen_excluded
                # or record_id not in man_prep_recs
            ):
                records[record_id][  # pylint: disable=colrev-direct-status-assign
                    Fields.STATUS
                ] = RecordState.rev_prescreen_excluded
                self.review_manager.logger.info(
                    f" {Colors.RED}{record_id}".ljust(46)
                    + "md_needs_manual_preparation →  rev_prescreen_excluded"
                    + f"{Colors.END}"
                )

        for record_id, man_prepped_record_dict in man_prep_recs.items():
            if record_id not in records:
                print(f"ID no longer in records: {record_id}")
                continue
            original_record = colrev.record.record.Record(
                records[man_prepped_record_dict[Fields.ID]]
            )
            self._import_record(
                man_prepped_record_dict=man_prepped_record_dict,
                original_record=original_record,
                imported_records=imported_records,
            )

        self.review_manager.dataset.save_records_dict(records)
        self.review_manager.dataset.create_commit(msg="Prep-man (ExportManPrep)")
        self.review_manager.dataset.set_ids(selected_ids=imported_records)
        self.review_manager.dataset.create_commit(msg="Set IDs")

    def _print_export_prep_man_instructions(self) -> None:
        print("Created two files:")
        if platform.system() == "Windows":
            print(
                f" - {self.prep_man_xlsx_path.relative_to(self.review_manager.path)}  (EXCEL file)"
            )
        else:
            print(
                f" - {self.prep_man_csv_path.relative_to(self.review_manager.path)}  (CSV file)"
            )
        print(
            f" - {self.prep_man_bib_path.relative_to(self.review_manager.path)}       (BIB file)"
        )
        print()
        print("To prepare records:")
        print("- check the defect codes in the CSV file")
        print(
            "- edit the BIB file and change the fields "
            "(e.g., add missing volume/number or remove all-caps)"
        )
        print(
            "- if a record should be exluded in the prescreen, simply remove it from the BIB file"
        )
        print()

        print(f"Once completed, run {Colors.ORANGE}colrev prep-man{Colors.END} again.")

    def prepare_manual(self, records: dict) -> dict:
        """Prepare records manually by extracting the subset of records to a separate BiBTex file"""

        if not self.prep_man_bib_path.is_file():
            self._create_info_dataframe(records=records)
            self._export_prep_man(records=records)
            self._print_export_prep_man_instructions()
        else:
            selected_path = self.prep_man_bib_path.relative_to(self.review_manager.path)
            if input(f"Import changes from {selected_path} [y,n]?") == "y":
                self._import_prep_man()

        return records



#!/usr/bin/env python3
"""Functionality for record ID setting."""
from __future__ import annotations

import itertools
import logging
import re
import string
import typing

from tqdm import tqdm

import colrev.env.utils
import colrev.exceptions as colrev_exceptions
import colrev.loader.bib
import colrev.loader.load_utils
import colrev.process.operation
import colrev.record.record
from colrev.constants import Fields
from colrev.constants import FieldValues
from colrev.constants import IDPattern
from colrev.constants import RecordState

# pylint: disable=too-few-public-methods


class IDSetter:
    """The IDSetter class"""

    def __init__(
        self,
        *,
        id_pattern: IDPattern,
        skip_local_index: bool = False,
        logger: logging.Logger = logging.getLogger(__name__),
    ) -> None:

        self.id_pattern = id_pattern
        self.skip_local_index = skip_local_index
        if not self.skip_local_index:
            self.local_index = colrev.env.local_index.LocalIndex()
        self.logger = logger

    def _get_author_last_names(self, record_dict: dict) -> list:
        if record_dict.get(Fields.AUTHOR, record_dict.get(Fields.EDITOR, "")) != "":
            authors_string = record_dict.get(
                Fields.AUTHOR, record_dict.get(Fields.EDITOR, FieldValues.ANONYMOUS)
            )
            authors = colrev.record.record_prep.PrepRecord.format_author_field(
                authors_string
            ).split(" and ")
        else:
            authors = [FieldValues.ANONYMOUS]

        # Use family names
        for i, author in enumerate(authors):
            if "," in author:
                authors[i] = author.split(",", maxsplit=1)[0]
            else:
                authors[i] = author.split(" ", maxsplit=1)[0]

        return authors

    def _generate_id_from_pattern(self, record_dict: dict) -> str:
        authors = self._get_author_last_names(record_dict)

        if IDPattern.first_author_year == self.id_pattern:
            temp_id = f'{authors[0]}{str(record_dict.get(Fields.YEAR, "NoYear"))}'
        elif IDPattern.three_authors_year == self.id_pattern:
            temp_id = ""
            indices = len(authors)
            if len(authors) > 3:
                indices = 3
            for ind in range(0, indices):
                temp_id = temp_id + f"{authors[ind]}"
            if len(authors) > 3:
                temp_id = temp_id + "EtAl"
            temp_id = temp_id + str(record_dict.get(Fields.YEAR, "NoYear"))

        # Replace special characters
        # (because IDs may be used as file names)
        temp_id = colrev.env.utils.remove_accents(temp_id)
        temp_id = temp_id.replace(" ", "")
        temp_id = re.sub(r"\(.*\)", "", temp_id)
        temp_id = re.sub("[^0-9a-zA-Z ]+", "", temp_id)
        if temp_id.isupper():  # pragma: no cover
            temp_id = temp_id.capitalize()
        return temp_id

    def _make_id_unique(
        self,
        temp_id: str,
        *,
        existing_ids: list,
    ) -> str:
        """Get the next unique ID"""

        order = 0
        letters = list(string.ascii_lowercase)
        next_unique_id = temp_id
        appends: list = []
        while next_unique_id.lower() in [i.lower() for i in existing_ids]:
            if len(appends) == 0:
                order += 1
                appends = list(itertools.product(letters, repeat=order))
            next_unique_id = temp_id + "".join(list(appends.pop(0)))
        temp_id = next_unique_id
        return temp_id

    def _generate_id(
        self,
        record_dict: dict,
        *,
        existing_ids: typing.Optional[list] = None,
    ) -> str:
        """Generate a blacklist to avoid setting duplicate IDs"""

        if self.skip_local_index:
            temp_id = self._generate_id_from_pattern(record_dict)
        else:
            try:
                retrieved_record = self.local_index.retrieve(record_dict)
                temp_id = retrieved_record.data[Fields.ID]  # pragma: no cover
            except (
                colrev_exceptions.RecordNotInIndexException,
                colrev_exceptions.NotEnoughDataToIdentifyException,
            ):
                temp_id = self._generate_id_from_pattern(record_dict)

        if existing_ids:
            temp_id = self._make_id_unique(
                temp_id,
                existing_ids=existing_ids,
            )

        return temp_id

    def set_ids(
        self, records: dict, *, selected_ids: typing.Optional[list] = None
    ) -> dict:
        """Set the IDs for the records in the dataset"""

        id_list = list(records.keys())

        for record_id in tqdm(list(records.keys())):
            record_dict = records[record_id]
            if selected_ids is not None:
                if record_id not in selected_ids:  # pragma: no cover
                    continue
            elif record_dict[Fields.STATUS] not in [
                RecordState.md_imported,
                RecordState.md_prepared,
            ]:
                continue
            old_id = record_id

            temp_stat = record_dict[Fields.STATUS]
            if selected_ids:
                record = colrev.record.record.Record(record_dict)
                record.set_status(RecordState.md_prepared)

            new_id = old_id
            # Only change IDs that are before md_processed
            if record_dict[Fields.STATUS] not in RecordState.get_post_x_states(
                state=RecordState.md_processed
            ):
                new_id = self._generate_id(
                    record_dict,
                    existing_ids=[x for x in id_list if x != record_id],
                )

            if selected_ids:
                record = colrev.record.record.Record(record_dict)
                record.set_status(temp_stat)

            self._update_id(
                records,
                id_list=id_list,
                record_dict=record_dict,
                old_id=old_id,
                new_id=new_id,
            )

        return records

    # pylint: disable=too-many-arguments
    def _update_id(
        self,
        records: dict,
        *,
        id_list: list,
        record_dict: dict,
        old_id: str,
        new_id: str,
    ) -> None:
        id_list.append(new_id)
        if old_id != new_id:
            # We need to insert the a new element into records
            # to make sure that the IDs are actually saved
            record_dict.update(ID=new_id)
            records[new_id] = record_dict
            del records[old_id]
            self.logger.info(f"set_ids({old_id}) to {new_id}")
            if old_id in id_list:
                id_list.remove(old_id)



#! /usr/bin/env python
"""Genral polishing rules"""
from __future__ import annotations

import re

import zope.interface
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields


# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.PrepInterface)
class GeneralPolishPrep:
    """Prepares records by applying polishing rules"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings

    ci_supported: bool = Field(default=True)

    source_correction_hint = "check with the developer"
    always_apply_changes = False

    frequent_acronyms = [
        "MIS",
        "GSS",
        "DSS",
        "ICIS",
        "ECIS",
        "AMCIS",
        "PACIS",
        "BI",
        "CRM",
        "CIO",
        "CEO",
        "ERP",
        "ICT",
        "Twitter",
        "Facebook",
        "Wikipedia",
        "B2B",
        "C2C",
    ]

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)

    def prepare(
        self, record: colrev.record.record_prep.PrepRecord
    ) -> colrev.record.record.Record:
        """Prepare the record by applying polishing rules"""

        if Fields.TITLE in record.data:
            acronyms = [
                x
                for x in self.frequent_acronyms
                if x.lower() in record.data[Fields.TITLE].lower().split()
            ]
            for acronym in acronyms:
                record.data[Fields.TITLE] = re.sub(
                    acronym, acronym, record.data[Fields.TITLE], flags=re.I
                )

        return record



#!/usr/bin/env python3
"""Create and apply record corrections in source repositories."""
from __future__ import annotations

import json
import typing
from pathlib import Path

from dictdiffer import diff

from colrev.constants import Fields

if typing.TYPE_CHECKING:  # pragma: no cover
    import colrev.review_manager

# pylint: disable=too-few-public-methods


class Corrections:
    """Handling corrections of metadata"""

    # pylint: disable=duplicate-code
    essential_md_keys = [
        Fields.TITLE,
        Fields.AUTHOR,
        Fields.JOURNAL,
        Fields.YEAR,
        Fields.BOOKTITLE,
        Fields.NUMBER,
        Fields.VOLUME,
        Fields.AUTHOR,
        Fields.DOI,
        Fields.ORIGIN,  # Note : for merges
    ]

    keys_to_ignore = [
        Fields.ID,
        Fields.SCREENING_CRITERIA,
        Fields.STATUS,
        "source_url",
        "metadata_source_repository_paths",
        "grobid-version",
        "colrev_pdf_id",
        Fields.FILE,
        Fields.ORIGIN,
        Fields.D_PROV,
        Fields.MD_PROV,
        Fields.SEMANTIC_SCHOLAR_ID,
        Fields.CITED_BY,
        Fields.ABSTRACT,
        Fields.PAGES,
    ]

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
    ) -> None:
        self.review_manager = review_manager
        self.corrections_path = self.review_manager.paths.corrections
        self.corrections_path.mkdir(exist_ok=True)

    def _record_corrected(self, *, prior_r: dict, record_dict: dict) -> bool:
        return not all(
            prior_r.get(k, "NA") == record_dict.get(k, "NA")
            for k in self.essential_md_keys
        )

    def _prep_for_change_item_creation(
        self, *, original_record: dict, corrected_record: dict
    ) -> None:
        # Cast to string for persistence
        original_record = {k: str(v) for k, v in original_record.items()}
        corrected_record = {k: str(v) for k, v in corrected_record.items()}

        # Note : removing the fields is a temporary fix
        # because the subsetting of change_items does not seem to
        # work properly
        keys_to_drop = [Fields.PAGES, Fields.STATUS]
        for k in keys_to_drop:
            original_record.pop(k, None)
            corrected_record.pop(k, None)

    def _get_selected_change_items(
        self, original_record: dict, corrected_record: dict
    ) -> list:
        changes = diff(original_record, corrected_record)
        selected_change_items = []
        for change_item in list(changes):
            change_type, key, val = change_item

            if not isinstance(key, str):
                continue

            if change_type != "add" and key == "":
                continue

            if key.split(".")[0] in self.keys_to_ignore:
                continue

            if change_type == "add":
                for add_item in val:
                    add_item_key, add_item_val = add_item
                    if not isinstance(add_item_key, str):
                        break
                    if add_item_key.split(".")[0] in self.keys_to_ignore:
                        break
                    selected_change_items.append(
                        ("add", "", [(add_item_key, add_item_val)])
                    )

            elif change_type == "change":
                selected_change_items.append(change_item)
        return selected_change_items

    def _create_change_item(
        self,
        *,
        original_record: dict,
        corrected_record: dict,
    ) -> None:

        self._prep_for_change_item_creation(
            original_record=original_record,
            corrected_record=corrected_record,
        )

        selected_change_items = self._get_selected_change_items(
            original_record, corrected_record
        )

        if len(selected_change_items) == 0:
            return

        if len(corrected_record.get(Fields.ORIGIN, [])) > len(
            original_record.get(Fields.ORIGIN, [])
        ):
            if (
                Fields.DBLP_KEY in corrected_record
                and Fields.DBLP_KEY in original_record
            ):
                if (
                    corrected_record[Fields.DBLP_KEY]
                    != original_record[Fields.DBLP_KEY]
                ):
                    selected_change_items = {  # type: ignore
                        "merge": [
                            corrected_record[Fields.DBLP_KEY],
                            original_record[Fields.DBLP_KEY],
                        ]
                    }
            # else:
            #     selected_change_items = {
            #         "merge": [
            #             corrected_record[Fields.ID],
            #             original_record[Fields.ID],
            #         ]
            #     }

        # cover non-masterdata corrections
        if Fields.MD_PROV not in original_record:
            return

        dict_to_save = {
            "original_record": {
                k: v for k, v in original_record.items() if k not in [Fields.STATUS]
            },
            "changes": selected_change_items,
        }

        filepath = self.corrections_path / Path(f"{corrected_record['ID']}.json")

        with open(filepath, "w", encoding="utf8") as corrections_file:
            json.dump(dict_to_save, corrections_file, indent=4)

    def check_corrections_of_records(self) -> None:
        """Check for corrections of records"""

        # to test run
        # colrev-hooks-report .report.log

        records = self.review_manager.dataset.load_records_dict()
        prior_records_dict = next(
            self.review_manager.dataset.load_records_from_history(), {}
        )
        for record_dict in records.values():
            # identify curated records for which essential metadata is changed
            record_prior = [
                x
                for x in prior_records_dict.values()
                if any(y in record_dict[Fields.ORIGIN] for y in x[Fields.ORIGIN])
            ]

            if len(record_prior) == 0:
                self.review_manager.logger.debug("No prior records found")
                continue

            for prior_r in record_prior:
                if self._record_corrected(prior_r=prior_r, record_dict=record_dict):
                    corrected_record = record_dict.copy()

                    self._create_change_item(
                        original_record=prior_r,
                        corrected_record=corrected_record,
                    )



#! /usr/bin/env python
"""SearchSource: backward search (based on PDFs and GROBID)"""
from __future__ import annotations

import json
import typing
from pathlib import Path

import inquirer
import pandas as pd
import requests
import zope.interface
from bib_dedupe.bib_dedupe import block
from bib_dedupe.bib_dedupe import cluster
from bib_dedupe.bib_dedupe import match
from bib_dedupe.bib_dedupe import prep
from bib_dedupe.merge import merge
from pydantic import Field
from rapidfuzz import fuzz
from tqdm import tqdm

import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
import colrev.record.record_prep
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields
from colrev.constants import RecordState
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType
from colrev.packages.crossref.src import crossref_api

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class BackwardSearchSource:
    """Backward search extracting references from PDFs using GROBID
    Scope: all included papers with colrev_status in (rev_included, rev_synthesized)
    """

    _api_url = "https://opencitations.net/index/coci/api/v1/references/"

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings
    endpoint = "colrev.pdf_backward_search"
    source_identifier = Fields.ID
    search_types = [SearchType.BACKWARD_SEARCH]

    ci_supported: bool = Field(default=False)
    heuristic_status = SearchSourceHeuristicStatus.supported

    def __init__(
        self, *, source_operation: colrev.process.operation.Operation, settings: dict
    ) -> None:
        self.review_manager = source_operation.review_manager
        if "min_intext_citations" not in settings["search_parameters"]:
            settings["search_parameters"]["min_intext_citations"] = 3

        self.search_source = self.settings_class(**settings)
        # Do not run in continuous-integration environment
        if not self.review_manager.in_ci_environment():
            self.grobid_service = self.review_manager.get_grobid_service()
            self.grobid_service.start()

        self.crossref_api = crossref_api.CrossrefAPI(params={})

    @classmethod
    def get_default_source(cls) -> colrev.settings.SearchSource:
        """Get the default SearchSource settings"""

        return colrev.settings.SearchSource(
            endpoint="colrev.pdf_backward_search",
            filename=Path("data/search/pdf_backward_search.bib"),
            search_type=SearchType.BACKWARD_SEARCH,
            search_parameters={
                "scope": {"colrev_status": "rev_included|rev_synthesized"},
                "min_intext_citations": 3,
            },
            comment="",
        )

    def _validate_source(self) -> None:
        """Validate the SearchSource (parameters etc.)"""
        source = self.search_source
        self.review_manager.logger.debug(f"Validate SearchSource {source.filename}")

        assert source.search_type == SearchType.BACKWARD_SEARCH

        if "scope" not in source.search_parameters:
            raise colrev_exceptions.InvalidQueryException(
                "Scope required in the search_parameters"
            )

        if source.search_parameters["scope"].get("file", "") == "paper.md":
            pass
        else:
            if (
                source.search_parameters["scope"]["colrev_status"]
                != "rev_included|rev_synthesized"
            ):
                raise colrev_exceptions.InvalidQueryException(
                    "search_parameters/scope/colrev_status must be rev_included|rev_synthesized"
                )

        self.review_manager.logger.debug(f"SearchSource {source.filename} validated")

    def _bw_search_condition(self, *, record: dict) -> bool:
        # rev_included/rev_synthesized required, but record not in rev_included/rev_synthesized
        if (
            "colrev_status" in self.search_source.search_parameters["scope"]
            and self.search_source.search_parameters["scope"]["colrev_status"]
            == "rev_included|rev_synthesized"
            and record[Fields.STATUS]
            not in [
                RecordState.rev_included,
                RecordState.rev_synthesized,
            ]
        ):
            return False

        # Note: this is for peer_reviews
        if "file" in self.search_source.search_parameters["scope"]:
            if (
                self.search_source.search_parameters["scope"]["file"] == "paper.pdf"
            ) and "data/pdfs/paper.pdf" != record.get(Fields.FILE, ""):
                return False

        if not (self.review_manager.path / Path(record[Fields.FILE])).is_file():
            self.review_manager.logger.error(f"File not found for {record[Fields.ID]}")
            return False

        return True

    def _get_reference_records_from_open_citations(self, *, record_dict: dict) -> list:
        references = []

        url = f"{self._api_url}{record_dict['doi']}"
        # headers = {"authorization": "YOUR-OPENCITATIONS-ACCESS-TOKEN"}
        headers: typing.Dict[str, str] = {}
        ret = requests.get(url, headers=headers, timeout=300)
        try:
            items = json.loads(ret.text)

            for doi in [x["cited"] for x in items]:
                try:
                    retrieved_record = self.crossref_api.query_doi(
                        doi=doi,
                    )
                    # if not crossref_query_return:
                    #     raise colrev_exceptions.RecordNotFoundInPrepSourceException()
                    retrieved_record.data[Fields.ID] = retrieved_record.data[Fields.DOI]
                    references.append(retrieved_record.data)
                except (
                    colrev_exceptions.RecordNotFoundInPrepSourceException,
                    colrev_exceptions.RecordNotParsableException,
                ):
                    pass
        except json.decoder.JSONDecodeError:
            self.review_manager.logger.info(
                f"Error retrieving citations from Opencitations for {record_dict['ID']}"
            )

        return references

    def _get_similarity(
        self, *, record: colrev.record.record.Record, retrieved_record_dict: dict
    ) -> float:
        title_similarity = fuzz.partial_ratio(
            retrieved_record_dict.get(Fields.TITLE, "").lower(),
            record.data.get(Fields.TITLE, "").lower(),
        )
        container_similarity = fuzz.partial_ratio(
            colrev.record.record_prep.PrepRecord(retrieved_record_dict)
            .get_container_title()
            .lower(),
            record.get_container_title().lower(),
        )
        weights = [0.6, 0.4]
        similarities = [title_similarity, container_similarity]

        similarity = sum(similarities[g] * weights[g] for g in range(len(similarities)))
        return similarity

    @classmethod
    def _deduplicate_all_references(cls, all_references: dict) -> pd.DataFrame:
        print("Resolving entities in citation network")
        # Flatten the list of lists into a single list of references, including the record ID
        all_references_flat = [
            {**ref, "record_id": record_id}
            for record_id, refs in all_references.items()
            for ref in refs
        ]
        if len(all_references_flat) == 0:
            return pd.DataFrame()

        for ref in all_references_flat:
            ref[Fields.ID] = ref["record_id"] + "_backward_search_" + ref[Fields.ID]

        # Create a DataFrame from the flattened list of references
        df_all_references = pd.DataFrame(all_references_flat)
        df_all_references["nr_references"] = 1

        def merge_into_list(values: list) -> str:
            """Concatenate all values into a single string, separated by commas."""
            merged_string = ",".join(str(value) for value in values)
            return merged_string

        def sum_nr_references(values: list) -> int:
            """Sum all integers in a list."""
            return sum(values)

        records_df = prep(df_all_references, verbosity_level=0)
        deduplication_pairs = block(records_df, verbosity_level=0)
        matched_df = match(deduplication_pairs, verbosity_level=0)
        duplicate_id_sets = cluster(matched_df, verbosity_level=0)
        df_all_references = merge(
            df_all_references,
            duplicate_id_sets=duplicate_id_sets,
            merge_functions={
                Fields.NR_INTEXT_CITATIONS: merge_into_list,
                "nr_references": sum_nr_references,
            },
        )
        df_all_references.rename(columns={"origin": "bw_search_origins"}, inplace=True)
        return df_all_references

    @classmethod
    def _export_crosstab_thresholds(cls, df_all_references: pd.DataFrame) -> None:
        cross_tabulated_data: typing.Dict[int, dict] = {}

        for in_text_citation_threshold in range(1, 20):
            for ref_freq_threshold in range(1, 20):

                df_all_references["meets_criteria"] = df_all_references[
                    Fields.NR_INTEXT_CITATIONS
                ].apply(
                    lambda x, threshold=in_text_citation_threshold, rft=ref_freq_threshold: len(
                        [
                            citation
                            for citation in x.split(",")
                            if int(citation) >= threshold
                        ]
                    )
                    >= rft
                )

                total_references = df_all_references["meets_criteria"].sum()

                if ref_freq_threshold not in cross_tabulated_data:
                    cross_tabulated_data[ref_freq_threshold] = {}
                cross_tabulated_data[ref_freq_threshold][
                    in_text_citation_threshold
                ] = total_references

        cross_tabulated_df = pd.DataFrame(cross_tabulated_data).T
        cross_tabulated_df.reset_index(inplace=True)
        cross_tabulated_df.columns = ["ref_freq_threshold"] + [
            "in_text_citation_threshold_" + str(i) for i in range(1, 20)
        ]
        cross_tabulated_df.to_csv(
            "cross_tabulated_evaluation.csv", header=True, index=False
        )
        print("Exported to cross_tabulated_evaluation.csv.")

    def _complement_with_open_citations_data(
        self,
        *,
        pdf_backward_search_feed: colrev.ops.search_api_feed.SearchAPIFeed,
        records: dict,
    ) -> None:
        self.review_manager.logger.info("Comparing records with open-citations data")

        for feed_record_dict in pdf_backward_search_feed.feed_records.values():
            parent_record = self._get_parent_record(feed_record_dict, records)
            if not parent_record:
                continue

            backward_references = self._get_reference_records_from_open_citations(
                record_dict=parent_record
            )
            self._update_feed_records_with_open_citations_data(
                feed_record_dict, backward_references
            )

    def _get_parent_record(self, feed_record_dict: dict, records: dict) -> dict:
        bw_search_origin = feed_record_dict["bw_search_origins"].split(";")[0]
        parent_record_id = bw_search_origin[
            : bw_search_origin.find("_backward_search_")
        ]
        return records.get(parent_record_id, {})

    def _update_feed_records_with_open_citations_data(
        self, feed_record_dict: dict, backward_references: list
    ) -> None:
        feed_record = colrev.record.record.Record(feed_record_dict)
        max_similarity, best_match = self._find_best_match(
            feed_record, backward_references
        )

        if max_similarity > 0.9:
            feed_record.data.update(**best_match)
            self.review_manager.logger.info(
                f"Updated record {feed_record.data[Fields.ID]} with OpenCitations data."
            )

    def _find_best_match(
        self, feed_record: colrev.record.record.Record, backward_references: list
    ) -> tuple:
        max_similarity = 0.0
        best_match = None
        for backward_reference in backward_references:
            similarity = self._get_similarity(
                record=feed_record, retrieved_record_dict=backward_reference
            )
            if similarity > max_similarity:
                best_match = backward_reference
                max_similarity = similarity
        return max_similarity, best_match

    def _get_new_record(
        self,
        *,
        item: dict,
        pdf_backward_search_feed: colrev.ops.search_api_feed.SearchAPIFeed,
    ) -> colrev.record.record.Record:
        item[Fields.ID] = "new"
        # Note: multiple source_identifiers are merged in origin field.
        for feed_record in pdf_backward_search_feed.feed_records.values():
            feed_record_origins = feed_record["bw_search_origins"].split(",")
            new_record_origins = item["bw_search_origins"].split(",")
            if set(feed_record_origins) & set(new_record_origins):
                item[Fields.ID] = feed_record[Fields.ID]
                break

        if item[Fields.ID] == "new":
            max_id = max(
                [
                    int(r[Fields.ID])
                    for r in pdf_backward_search_feed.feed_records.values()
                ]
                or [0]
            )
            item[Fields.ID] = str(max_id + 1).rjust(6, "0")

        fields_to_drop = [
            "meets_criteria",
            "record_id",
            "nr_references",
            "nr_intext_citations",
            "bwsearch_ref",
        ]
        for field in fields_to_drop:
            if field in item:
                del item[field]
        item = {
            k: v for k, v in item.items() if not pd.isna(v)
        }  # Drop fields where value is NaN
        return colrev.record.record.Record(item)

    def search(self, rerun: bool) -> None:
        """Run a search of PDFs (backward search based on GROBID)"""

        self._validate_source()

        # Do not run in continuous-integration environment
        if self.review_manager.in_ci_environment():
            raise colrev_exceptions.SearchNotAutomated(
                "PDF Backward Search not automated."
            )

        records = self.review_manager.dataset.load_records_dict()

        if not records:
            self.review_manager.logger.info(
                "No records imported. Cannot run backward search yet."
            )
            return
        min_intext_citations = self.search_source.search_parameters[
            "min_intext_citations"
        ]
        self.review_manager.logger.info(
            f"Set min_intext_citations={min_intext_citations}"
        )
        nr_references_threshold = self.search_source.search_parameters.get(
            "min_ref_freq", 1
        )
        self.review_manager.logger.info(
            f"Set nr_references_threshold={nr_references_threshold}"
        )

        selected_records = {
            rid: record
            for rid, record in records.items()
            if self._bw_search_condition(record=record)
        }

        all_references = self._get_all_references(
            selected_records=selected_records, review_manager=self.review_manager
        )

        df_all_references = self._deduplicate_all_references(all_references)

        df_all_references["meets_criteria"] = df_all_references[
            Fields.NR_INTEXT_CITATIONS
        ].apply(
            lambda x: len(
                [
                    citation
                    for citation in x.split(",")
                    if int(citation) >= int(min_intext_citations)
                ]
            )
            >= int(nr_references_threshold)
        )

        selected_references = df_all_references[df_all_references["meets_criteria"]]
        pdf_backward_search_feed = self.search_source.get_api_feed(
            review_manager=self.review_manager,
            source_identifier=self.source_identifier,
            update_only=(not rerun),
        )

        for item in selected_references.to_dict(orient="records"):
            new_record = self._get_new_record(
                item=item, pdf_backward_search_feed=pdf_backward_search_feed
            )

            pdf_backward_search_feed.add_update_record(
                retrieved_record=new_record,
            )

        self._complement_with_open_citations_data(
            pdf_backward_search_feed=pdf_backward_search_feed, records=records
        )

        pdf_backward_search_feed.save()

        if self.review_manager.dataset.has_record_changes():
            self.review_manager.dataset.create_commit(
                msg="Backward search", script_call="colrev search"
            )

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for PDF backward searches (GROBID)"""

        result = {"confidence": 0.0}
        if str(filename).endswith("_ref_list.pdf"):
            result["confidence"] = 1.0
            return result
        return result

    @classmethod
    def _get_all_references(
        cls,
        *,
        selected_records: dict,
        review_manager: colrev.review_manager.ReviewManager,
    ) -> dict:

        all_references = {}

        for record in tqdm(selected_records.values()):
            try:

                review_manager.logger.info(
                    f" run backward search for {record[Fields.ID]}"
                )

                pdf_path = review_manager.path / Path(record[Fields.FILE])
                tei = review_manager.get_tei(
                    pdf_path=pdf_path,
                    tei_path=colrev.record.record.Record(record).get_tei_filename(),
                )

                references = tei.get_references(add_intext_citation_count=True)
                for reference in references:
                    if "tei_id" in reference:
                        del reference["tei_id"]
                    reference["bwsearch_ref"] = (
                        record[Fields.ID] + "_backward_search_" + reference[Fields.ID]
                    )

                all_references[record[Fields.ID]] = references

            except colrev_exceptions.TEIException:
                review_manager.logger.info("Error accessing TEI")
            except KeyError as exc:
                review_manager.logger.info(exc)
        return all_references

    @classmethod
    def _get_settings_from_ui(
        cls, *, params: dict, review_manager: colrev.review_manager.ReviewManager
    ) -> None:

        question = "Do you want to create an overview for min_ref_freq and min_intext_citations?"
        create_overview = inquirer.confirm(question)

        if not create_overview:
            return

        print(
            "Calculating sample sizes for parameters min_ref_freq and min_intext_citations"
        )
        # Assuming there's a function to create an overview, it would be called here.
        # For example: cls.create_min_intext_citations_overview(params)

        records = review_manager.dataset.load_records_dict()
        selected_records = {
            record_id: record
            for record_id, record in records.items()
            if record[Fields.STATUS]
            in [
                RecordState.rev_included,
                RecordState.rev_synthesized,
            ]
        }

        all_references = cls._get_all_references(
            selected_records=selected_records, review_manager=review_manager
        )

        df_all_references = cls._deduplicate_all_references(all_references)

        cls._export_crosstab_thresholds(df_all_references)

        questions = [
            inquirer.Text(
                "in_text_citation_threshold", message="Enter in-text citation threshold"
            ),
            inquirer.Text(
                "ref_freq_threshold", message="Enter reference frequency threshold"
            ),
        ]
        answers = inquirer.prompt(questions)

        try:
            params["min_intext_citations"] = int(answers["in_text_citation_threshold"])
            params["min_ref_freq"] = int(answers["ref_freq_threshold"])
        except ValueError:
            print("Both thresholds must be integers. Please try again.")
            return

    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint (based on query provided to colrev search --add )"""

        params_dict = {}
        if params:
            for item in params.split(";"):
                key, value = item.split("=")
                params_dict[key] = value

        if "min_intext_citations" not in params_dict:
            cls._get_settings_from_ui(
                params=params_dict, review_manager=operation.review_manager
            )
        else:
            assert params_dict["min_intext_citations"].isdigit()
            assert params_dict["min_ref_freq"].isdigit()

        search_source = cls.get_default_source()
        if "min_intext_citations" in params_dict:
            search_source.search_parameters["min_intext_citations"] = int(
                params_dict["min_intext_citations"]
            )
        if "min_ref_freq" in params_dict:
            search_source.search_parameters["min_ref_freq"] = int(
                params_dict["min_ref_freq"]
            )

        operation.add_source_and_search(search_source)
        return search_source

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            records = colrev.loader.load_utils.load(
                filename=self.search_source.filename,
                logger=self.review_manager.logger,
            )
            for record_dict in records.values():
                record_dict.pop("bw_search_origins")
            return records

        raise NotImplementedError

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.record.Record:
        """Not implemented"""
        return record

    def prepare(
        self,
        record: colrev.record.record_prep.PrepRecord,
        source: colrev.settings.SearchSource,
    ) -> colrev.record.record.Record:
        """Source-specific preparation for PDF backward searches (GROBID)"""

        record.format_if_mostly_upper(Fields.TITLE, case="sentence")
        record.format_if_mostly_upper(Fields.JOURNAL, case=Fields.TITLE)
        record.format_if_mostly_upper(Fields.BOOKTITLE, case=Fields.TITLE)
        record.format_if_mostly_upper(Fields.AUTHOR, case=Fields.TITLE)

        if (
            "multimedia appendix"
            in record.data.get(Fields.TITLE, "").lower()
            + record.data.get(Fields.JOURNAL, "").lower()
        ):
            record.prescreen_exclude(reason="grobid-error")

        if (
            record.data[Fields.ENTRYTYPE] == ENTRYTYPES.MISC
            and Fields.PUBLISHER in record.data
        ):
            record.data[Fields.ENTRYTYPE] = ENTRYTYPES.BOOK

        return record



#! /usr/bin/env python
"""Removal of broken IDs as a prep operation"""
from __future__ import annotations

import zope.interface
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import DefectCodes
from colrev.constants import Fields


# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.PrepInterface)
class RemoveBrokenIDPrep:
    """Prepares records by removing invalid IDs DOIs/ISBNs"""

    ci_supported: bool = Field(default=True)

    source_correction_hint = "check with the developer"
    always_apply_changes = True

    # check_status: relies on crossref / openlibrary connectors!
    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,
        settings: dict,
    ) -> None:
        self.settings = colrev.package_manager.package_settings.DefaultSettings(
            **settings
        )
        self.prep_operation = prep_operation
        self.review_manager = prep_operation.review_manager

    def prepare(
        self, record: colrev.record.record_prep.PrepRecord
    ) -> colrev.record.record.Record:
        """Prepare the record by removing broken IDs (invalid DOIs/ISBNs)"""

        if self.prep_operation.polish and not self.review_manager.force_mode:
            return record

        if DefectCodes.DOI_NOT_MATCHING_PATTERN in record.get_field_provenance_notes(
            Fields.DOI
        ):
            record.remove_field(key=Fields.DOI)

        if DefectCodes.ISBN_NOT_MATCHING_PATTERN in record.get_field_provenance_notes(
            Fields.ISBN
        ):
            record.remove_field(key=Fields.ISBN)

        return record
